import pandas as pd
import numpy as np

from abc import abstractmethod


class ProtectedAttribute():
    """
    This class retrieves all informations on a protected attribute on a specific dataset.

    Parent class of DatasetProtectedAttribute and ModelProtectedAttribute.

    Attributes
    ----------
    name: str
        Attribute name (the column's name in original dataframe)
    target:
        Name of the label (or target) column
    favorable_label:
        A specific value that is has an advantage over the other(s) values
    privileged_values: list
        List with privileged values inside the column (e.g. ['Male'] for 'gender' attribute)  
    unprivileged_values: list
        List with unprivileged values inside the column (e.g. ['Female'] for 'gender' attribute)  
    values: pd.Series
        Serie with the same length than labels attribute with 0 or 1 values 
        1 if it correspond to a privileged value, 0 if not.
    labels: np.array 
        Array with original label values        
    crosstab: pd.DataFrame
        Crosstab of values and labels
    """

    def __init__(self, dataset, attr, privileged_values, favorable_label=None):
        """
        Parameters
        ----------
        dataset: pd.DataFrame
            Dataframe to inspect
        attr: str
            Name of the attribute to analyse (it has to be into dataset columns)
        privileged_values: list
            List with privileged values inside the column (e.g. ['Male'] for 'gender' attribute)  
        favorable_label:
            A specific value that is has an advantage over the other(s) values
        """
        self.name = attr
        self.target = dataset.target
        self.favorable_label = favorable_label
        self.privileged_values = privileged_values
        self.unprivileged_values = [
            v for v in dataset.df[attr].unique() if v not in privileged_values]

        self.values = pd.Series(
            np.where(dataset.df[attr].isin(privileged_values), 1, 0), name=attr).values
        self.labels = dataset.df[dataset.target].values
        self.crosstab = pd.crosstab(self.values, self.labels, margins=True)

        if 0 not in self.crosstab.index:
            self.crosstab[0] = 0
        if 1 not in self.crosstab.index:
            self.crosstab[1] = 0

    def to_frame(self):
        """
        Returns a dataframe with 2 columns : values in first column and 
        labels in the second one

        Returns
        -------
        pd.DataFrame
            formated dataframe
        """
        df = pd.DataFrame()
        df[self.name] = self.values
        df[self.target] = self.labels
        return df

    def get_privileged_values(self):
        return f'{self.name} = '+' or '.join([str(v) for v in self.privileged_values])

    def get_unprivileged_values(self):
        return f'{self.name} = '+' or '.join([str(v) for v in self.unprivileged_values])

    def num_instances(self, privileged=None):
        """
        Compute the number of instances, :math:`n`, in the dataset conditioned
        on protected attributes if necessary.

        Parameters
        ----------
        privileged (bool, default None): 
                Boolean prescribing whether to
                condition this metric on the `privileged_groups`, if `True`, or
                the `unprivileged_groups`, if `False`. Defaults to `None`
                meaning this metric is computed over the entire dataset.

        Returns
        -------
        int:
            Number of instances of all data if privileged is None, privileged
            values if privileged is True and unprivileged values if False.
        """
        if privileged == None:
            return self.crosstab.loc['All', 'All']
        elif privileged:
            return self.crosstab.loc[1, 'All']
        else:
            return self.crosstab.loc[0, 'All']

    def difference(self, metric_fun, target_value):
        """
        Compute difference of the metric for unprivileged and privileged groups.

        Parameters
        ----------
        metric_fun: function
            metric function that returns a number
        target_value:
            Specific value of the target

        Returns
        -------
        float:
            Difference of a metric for unprivileged and privileged groups.
        """
        return (metric_fun(target_value=target_value, privileged=False)
                - metric_fun(target_value=target_value, privileged=True))

    def ratio(self, metric_fun, target_value):
        """
        Compute ratio of the metric for unprivileged and privileged groups.

        Parameters
        ----------
        metric_fun: function
            metric function that returns a number
        target_values:
            Specific value of the target

        Returns
        -------
        float:
            Ratio of a metric for unprivileged and privileged groups.
        """
        return (metric_fun(target_value=target_value, privileged=False)
                / metric_fun(target_value=target_value, privileged=True))

    def compute_bias_metrics(self, metrics_dict):
        """
        Computes bias metrics using a metrics_dict which has 
        metric names on keys and the associated function as values.

        Parameters
        ----------
        metrics_dict: dict
            Dictionnary with metric name on keys and 
            metric function as value
        """
        if self.favorable_label is not None:
            target_values = [self.favorable_label]
        else:
            if type(self.labels) in [pd.Series, pd.DataFrame]:
                target_values = self.labels.unique()
            else:
                target_values = list(set(self.labels))

        metrics = pd.DataFrame()
        metrics.name = self.name

        for target_value in target_values:
            for metric, func in metrics_dict.items():
                val = func(target_value=target_value)
                metrics.loc[target_value, metric] = val

        self.metrics = metrics

    # Abstract methods

    @abstractmethod
    def get_freq(self):
        return None


import pandas as pd
import numpy as np
import scipy.stats as ss
from IPython.display import display, Markdown

import transparentai.datasets.datasets_plots as plots
import transparentai.utils as utils


class StructuredDataset():
    """
    Class to inspect a strutured (tabular) dataset based on a pandas DataFrame object.
    It could help you to explore your data, understand what's in it with plot functions.
    
    If target is None then display different plots without splitting it by target. 

    In case the target is a continuous numeric variable and you specify it with `target_regr`
    attribute then it convert the target column to a binary classification variable greater than
    the mean and lesser or equal.   

    It helps to have a better understanding of your data.

    Example
    -------
    For binary classification :

    >>> from transparentai.datasets import StructuredDataset, load_adult
    >>> adult = load_adult()
    >>> dataset = StructuredDataset(df=adult, target='income')

    For regression :

    >>> from transparentai.datasets import StructuredDataset, load_boston
    >>> boston = load_boston()
    >>> dataset = StructuredDataset(df=boston, target='MEDV', target_regr=True)

    With no target column :

    >>> from transparentai.datasets import StructuredDataset, load_adult
    >>> adult = load_adult()
    >>> dataset = StructuredDataset(df=adult)

    Attributes
    ----------
    df: pd.DataFrame
        Dataframe with appropriate dtypes
    target: str (optional)
        target column, if None then this StructuredDataset will not
        be abble to use bias insight
    target_mean: number (optional)
        If the target is for a regression model (continuous variable) 
        Mean of the target variable 
    orig_target_value: pd.Series (optional)
        If the target is for a regression model (continuous variable) 
        this attribute contains original target values.
    """
    orig_target_value = None
    target_mean = None

    def __init__(self, df, target=None, mean=None, target_regr=False):
        """
        Parameters
        ----------
        df: pd.DataFrame
            Dataframe with appropriate dtypes
        target: str
            target column, if None then this StructuredDataset will not
            be abble to use bias insight
        mean: number (optional)
            If the target is for a regression model (continuous variable) 
            Mean of the target variable 
        target_regr: bool
            Whether the target is for a regression model (continuous variable)
            or not.

        Raises
        ------
        TypeError:
            Must provide a pandas DataFrame representing the data (features, target)
        ValueError:
            target attribute has to be in the df columns   
        """
        if (df is None) & (type(df) is not pd.DataFrame):
            raise TypeError("Must provide a pandas DataFrame representing " +
                            "the data (features, target)")
        if (target is not None) and (target not in df.columns):
            raise ValueError('target attribute has to be in the df columns')

        df = df.copy()

        if target is not None:
            if target in df.select_dtypes('object').columns:
                df[target] = df[target].astype('category')

            elif (target in df.select_dtypes('number').columns) & (target_regr):
                self.target_mean = mean if mean is not None else np.mean(
                    df[target]).round(3)
                df, orig_val = utils.regression_to_classification(
                    df, target, self.target_mean)
                self.orig_target_value = orig_val

        self.df = df.copy()
        self.target = target

    def __str__(self):
        return str(self.df)

    def _reduce_df_nrows(self, nrows=None):
        """
        Returns a reduce version of the df attribute dataframe.

        It usefull if there is too much data.
        
        Parameters
        ----------
        nrows: int (optional)
            If None then returns the original dataset
            else returns a sample of the dataset
        Returns
        -------
        pd.DataFrame:
            Reduced df attribute dataframe
        """
        if nrows is not None:
            df = utils.reduce_df_nrows(df=self.df, nrows=nrows)
        else:
            df = self.df
        return df

    def plot_dataset_overview(self):
        """
        Displays an overview of the dataset :

        - Shape 
        - Head
        """
        display(Markdown(f'#### Dataset shape : {self.df.shape}'))
        display(Markdown('First 5 rows of the dataset : '))
        display(self.df.head())

    def plot_missing_values(self):
        """
        Displays a bar chart of missing values for columns which contains
        at least one missing value.
        """
        plots.plot_missing_values(df=self.df)

    def plot_one_numeric_variable(self, var, nrows=None):
        """
        Display multiple graphics of a numerical variable.

        Parameters
        ----------
        var: str
            Variable name of a numerical variable inside dataframe object 
        nrows: int (optional)       
            If not None reduce the data to a sample of nrows

        Raises
        ------
        TypeError:
            Must provide a valid variable name string
        ValueError:
            The variable must be in the columns of the data
        ValueError:
            The variable has to be a numerical variable inside the dataframe
        """
        if var is None:
            raise TypeError("Must provide a valid variable name string")
        if var not in self.df.columns:
            raise ValueError(
                'The variable was not found inside the columns data')
        if var not in self.df.select_dtypes('number').columns:
            raise ValueError(
                'The variable has to be a numerical variable inside the dataframe')

        df = self._reduce_df_nrows(nrows=nrows)
        
        if var == self.target:
            plots.plot_numerical_var(df=df, var=self.target) 
        else:
            plots.plot_numerical_var(df=df, var=var, target=self.target)

    def plot_one_categorical_variable(self, var, nrows=None):
        """
        Display multiple graphics of a categorical variable.

        Parameters
        ----------
        var: str
            Variable name of a categorical variable inside dataframe object   
        nrows: int (optional)       
            If not None reduce the data to a sample of nrows     

        Raises
        ------
        TypeError:
            Must provide a valid variable name string
        ValueError:
            The variable must be in the columns of the data
        ValueError:
            The variable has to be a categorical variable inside the dataframe
        """
        if var is None:
            raise TypeError("Must provide a valid variable name string")
        if var not in self.df.columns:
            raise ValueError(
                'The variable was not found inside the columns data')
        if var not in self.df.select_dtypes(['object', 'category']).columns:
            raise ValueError(
                'The variable has to be a categorical variable inside the dataframe')

        df = self._reduce_df_nrows(nrows=nrows)

        if var == self.target:
            plots.plot_categorical_var(df=df, var=self.target) 
        else:
            plots.plot_categorical_var(df=df, var=var, target=self.target)

    def plot_one_datetime_variable(self, var, nrows=None):
        """
        Display multiple graphics of a datetime variable.

        Parameters
        ----------
        var: str
            Variable name of a datetime variable inside dataframe object  
        nrows: int (optional)       
            If not None reduce the data to a sample of nrows      

        Raises
        ------
        TypeError:
            Must provide a valid variable name string
        ValueError:
            The variable must be in the columns of the data
        ValueError:
            The variable has to be a datetime variable inside the dataframe
        """
        if var is None:
            raise TypeError("Must provide a valid variable name string")
        if var not in self.df.columns:
            raise ValueError(
                'The variable was not found inside the columns data')
        if var not in self.df.select_dtypes('datetime').columns:
            raise ValueError(
                'The variable has to be a datetime variable inside the dataframe')

        df = self._reduce_df_nrows(nrows=nrows)

        plots.plot_datetime_var(df=df, var=var, target=self.target)

    def plot_variables(self, nrows=None):
        """
        Display multiple graphics of all differents variables.
        The goal of this function is to do less code and get insightfull 
        informations about the data.

        Data type handle : categorical, numerical, datetime

        Parameters
        ----------
        nrows: int (optional)       
            If not None reduce the data to a sample of nrows
        """
        cat_vars = self.df.select_dtypes(['object', 'category'])
        num_vars = self.df.select_dtypes('number')
        dat_vars = self.df.select_dtypes('datetime')

        if self.target is not None:
            df = self._reduce_df_nrows(nrows=nrows)
            display(Markdown('### Target variable'))

            if self.target in cat_vars.columns:
                plots.plot_categorical_var(df=df, var=self.target)
                cat_vars = cat_vars.drop(columns=self.target)

            elif self.target in num_vars.columns:
                plots.plot_numerical_var(df=df, var=self.target)
                num_vars = num_vars.drop(columns=self.target)

        if len(num_vars) > 0:
            display(Markdown('### Numerical variables'))
        for var in num_vars:
            display(Markdown(''))
            plots.display_meta_var(self.df, var)
            if len(self.df[var].unique()) <= 1:
                display(Markdown('Only one value.'))
                continue
            self.plot_one_numeric_variable(var=var, nrows=nrows)

        if len(cat_vars) > 0:
            display(Markdown('### Categorical variables'))
        for var in cat_vars:
            display(Markdown(''))
            plots.display_meta_var(self.df, var)
            if len(self.df[var].unique()) <= 1:
                display(Markdown('Only one value.'))
                continue
            self.plot_one_categorical_variable(var=var, nrows=nrows)

        if len(num_vars) > 0:
            display(Markdown('### Datetime variables'))
        for var in dat_vars:
            display(Markdown(''))
            plots.display_meta_var(self.df, var)
            if len(self.df[var].unique()) <= 1:
                display(Markdown('Only one value.'))
                continue
            self.plot_one_datetime_variable(var=var, nrows=nrows)

    def plot_two_numeric_variables(self, var1, var2, nrows=None):
        """
        Show two numerical variables relations with jointplot.

        Parameters
        ----------
        var1: str
            Column name that contains first numerical values
        var2: str
            Column name that contains second numerical values
        nrows: int (optional)       
            If not None reduce the data to a sample of nrows

        Raises
        ------
        TypeError:
            Must provide a valid variable name string
        ValueError:
            Both variables must be in the columns of the data
        ValueError:
            Both variables has to be numeric variables inside the dataframe
        """
        if var1 is None or var2 is None:
            raise TypeError("Must provide valid variables name string")
        if (var1 not in self.df.columns) or (var2 not in self.df.columns):
            raise ValueError(
                'At least one of the variables was not found inside the columns data')
        if (var1 not in self.df.select_dtypes('number').columns) or (var2 not in self.df.select_dtypes('number').columns):
            raise ValueError(
                'Both variables has to be a numerical variables inside the dataframe')

        df = self._reduce_df_nrows(nrows=nrows)

        plots.plot_numerical_jointplot(
            df=df, var1=var1, var2=var2, target=self.target)

    def plot_numeric_var_relation(self, nrows=None):
        """
        Show all numerical variables 2 by 2 with graphics understand their relation.
        If target is set, separate dataset for each target value.
        """
        num_vars = self.df.select_dtypes('number')
        num_vars = utils.remove_var_with_one_value(num_vars)

        cols = num_vars.columns.values
        var_combi = [tuple(sorted([v1, v2]))
                     for v1 in cols for v2 in cols if (v1 != v2) & (self.target not in [v1, v2])]
        var_combi = list(set(var_combi))

        for var1, var2 in var_combi:
            display(Markdown(''))
            display(Markdown(f'Joint plot for **{var1}** & **{var2}**'))
            self.plot_two_numeric_variables(var1=var1, var2=var2, nrows=nrows)

    def plot_one_cat_and_num_variables(self, var1, var2, nrows=None):
        """
        Show boxplots for a specific pair of categorical and numerical variables
        If target is set, separate dataset for each target value.

        Parameters
        ----------
        var1: str
            Column name that contains categorical values
        var2: str
            Column name that contains numerical values
        nrows: int (optional)       
            If not None reduce the data to a sample of nrows

        Raises
        ------
        TypeError:
            Must provide valid variables name string
        ValueError:
            Both variables must be in the columns of the data
        """
        if var1 is None or var2 is None:
            raise TypeError("Must provide valid variables name string")
        if (var1 not in self.df.columns) or (var2 not in self.df.columns):
            raise ValueError(
                'At least one of the variables was not found inside the columns data')

        df = self._reduce_df_nrows(nrows=nrows)
        
        if var1 == self.target:
            plots.plot_one_cat_and_num_variables(
                df=df, var1=var1, var2=var2)
        else:
            plots.plot_one_cat_and_num_variables(
                df=df, var1=var1, var2=var2, target=self.target)

    def plot_cat_and_num_variables(self, nrows=None):
        """
        Show boxplots for each pair of categorical and numerical variables
        If target is set, separate dataset for each target value.

        Parameters
        ----------
        nrows: int (optional)       
            If not None reduce the data to a sample of nrows
        """
        df = utils.remove_var_with_one_value(self.df)

        num_vars = df.select_dtypes('number')
        cat_vars = df.select_dtypes(['object', 'category'])

        var_combi = [(v1, v2) for v1 in num_vars.columns for v2 in cat_vars.columns if (
            v1 != v2) & (self.target not in [v1, v2])]

        if self.target is not None:
            for num_var in num_vars.columns:
                display(Markdown(''))
                display(
                    Markdown(f'Box plot for **{self.target}** & **{num_var}**'))
                self.plot_one_cat_and_num_variables(
                    var1=self.target, var2=num_var, nrows=nrows)

        for num_var, cat_var in var_combi:
            display(Markdown(''))
            display(Markdown(f'Box plot for **{cat_var}** & **{num_var}**'))
            self.plot_one_cat_and_num_variables(
                var1=cat_var, var2=num_var, nrows=nrows)

    def plot_correlations(self, nrows=None):
        """
        Show differents correlations matrix for 3 cases :
        
        - numerical to numerical (using Pearson coeff)
        - categorical to categorical (using Cramers V & Chi square)
        - numerical to categorical (discrete) (using Point Biserial)

        Parameters
        ----------
        nrows: int (optional)       
            If not None reduce the data to a sample of nrows
        """
        df = self._reduce_df_nrows(nrows=nrows)
        df = utils.remove_var_with_one_value(df)

        if self.orig_target_value is not None:
            df[self.target] = self.orig_target_value

        num_df = df.select_dtypes('number')
        cat_df = df.select_dtypes(['object', 'category'])
        num_vars = num_df.columns
        cat_vars = cat_df.columns

        ignore_cat_vars = list()
        for var in cat_vars:
            if cat_df[var].nunique() > 100:
                ignore_cat_vars.append(var)
        cat_vars = [v for v in cat_vars if v not in ignore_cat_vars]

        if len(ignore_cat_vars) > 0:
            print('Ignored categorical variables because there are more than 100 values :', ', '.join(
                ignore_cat_vars))

        if len(num_df) > 0:
            pearson_corr = num_df.corr()
            display(
                Markdown('#### Pearson correlation matrix for numerical variables'))
            plots.plot_correlation_matrix(pearson_corr, fname='pearson_corr.png')

        if len(cat_vars) > 0:
            var_combi = [tuple(sorted([v1, v2]))
                         for v1 in cat_vars for v2 in cat_vars if v1 != v2]
            var_combi = list(set(var_combi))

            cramers_v_corr = utils.init_corr_matrix(
                columns=cat_vars, index=cat_vars)

            for var1, var2 in var_combi:
                corr = utils.cramers_v(cat_df[var1], cat_df[var2])
                cramers_v_corr.loc[var1, var2] = corr
                cramers_v_corr.loc[var2, var1] = corr

            display(
                Markdown('#### Cramers V correlation matrix for categorical variables'))
            plots.plot_correlation_matrix(cramers_v_corr, fname='cramers_v_corr.png')

        if (len(cat_vars) > 0) and (len(num_df) > 0):
            data_encoded, _ = utils.encode_categorical_vars(df)

            var_combi = [(v1, v2)
                         for v1 in cat_vars for v2 in num_vars if v1 != v2]

            pbs_corr = utils.init_corr_matrix(
                columns=num_vars, index=cat_vars, fill_diag=0.)

            for cat_var, num_var in var_combi:
                tmp_df = data_encoded[[cat_var, num_var]].dropna()
                if len(tmp_df) == 0:
                    continue
                corr, p_value = ss.pointbiserialr(
                    tmp_df[cat_var], tmp_df[num_var])
                pbs_corr.loc[cat_var, num_var] = corr

            display(Markdown(
                '#### Point Biserial correlation matrix for numerical & categorical variables'))
            plots.plot_correlation_matrix(pbs_corr, fname='pointbiserialr_corr.png')



__all__ = [
    'str_to_file', 
    'remove_var_with_one_value',
    'encode_categorical_vars',
    'cramers_v',
    'init_corr_matrix',
    'regression_to_classification',
    'labelencoder_to_dict',
    'get_metric_goal',
    'save_dict_to_json',
    'reduce_df_nrows'
    ]
    
from transparentai.utils.utils import * 


import matplotlib.pyplot as plt
import seaborn as sns

import transparentai.utils as utils
import transparentai.plots as plots
        
    
def plot_bar_performance(orig_perf=None, new_perf=None, alert_threshold=None):
    """
    Plot performance in `Monitoring` class with one bar chart for each metric.

    If `orig_perf` is not None but `new_perf` is, then plots only original 
    performance metrics.
    If `orig_perf` is None but `new_perf` is not, then plots only new 
    performance metrics.
    If neither `orig_perf` and `new_perf` are not None, then plots
    comparison between original metrics and new ones.

    You can add a threshold to see if the metrics are good or bad.

    Parameters
    ----------
    orig_perf: dict (default: None)
        Dictionary with metrics in keys and 
        original metric values in values
    new_perf: dict (default: None)
        Dictionary with metrics in keys and 
        new metric values in values 
    alert_threshold: dict (default: None)
        Dictionary with metrics in keys and
        threshold values in values

    Raises
    ------
    AttributeError:
        At least one of `orig_perf` or `new_perf` has to be set.
    """
    if (orig_perf is None) & (new_perf is None):
        raise AttributeError('Either orig_perf or new_perf has to be set')
    
    perf, perf_2 = (orig_perf, new_perf) if orig_perf is not None else (new_perf, None)
    txt, txt_2 = ('Original', 'New') if orig_perf is not None else ('New', '')
    
    height = int((len(perf) + len(perf)%2) / 2)
    fig, ax = plt.subplots(figsize=(15, 4*height))
    labels = []
    
    i = 1
    for k, v in perf.items():
        ax = plt.subplot(int(f'{height}2{i}'))
        plots.plot_bar(ax, value=v, label=k+f' {txt}')
        
        if perf_2 is not None:
            if k in perf_2:
                plots.plot_bar(ax, value=perf_2[k], label=k+f' {txt_2}')
                
        
        if alert_threshold is not None:
            if k in alert_threshold:
                ax.axhline(y=alert_threshold[k], linewidth=2, color='red')
                
        ax.set_title(f'{k} metric', fontsize=16)
        i += 1
        
    
    if alert_threshold is not None:
        labels += ['Alert threshold']
    
    labels += [txt] if perf_2 is None else [txt, txt_2]
    fig.legend(labels=labels, fontsize=16)
    
    fig.suptitle('Performance bar plot', fontsize=18)
    plots.plot_or_save(fname='monitoring_bar_performance_plot')
        
        
def plot_gauge_bias(attr=None, target_value=None, orig_bias=None, new_bias=None):
    """
    Plot bias in `Monitoring` class with one gauge plot for each metric.
    
    If `attr` and `target_value` it reduces the number of plots 
    Example : if you have 2 target values (e.g. 0 and 1) and different protected
    attributes (e.g. gender and marital status) then you can say that it will
    only plots for gender and target value = 1 

    If `orig_bias` is not None but `new_bias` is, then plots only original 
    performance metrics.
    If `orig_bias` is None but `new_bias` is not, then plots only new 
    performance metrics.
    If neither `orig_bias` and `new_bias` are not None, then plots
    comparison between original metrics and new ones.

    Parameters
    ----------
    attr: str (default: None)
        Name of the attribute to analyse
    target_value: (default: None)
        Specific label value 
    orig_bias: dict (default: None)
        Dictionary with metrics in keys and 
        original metric values in values
    new_bias: dict (default: None)
        Dictionary with metrics in keys and 
        new metric values in values 

    Raises
    ------
    AttributeError:
        At least one of `orig_bias` or `new_bias` has to be set.
    """
    if (orig_bias is None) & (new_bias is None):
        raise AttributeError('Either orig_bias or new_bias has to be set')
    
    bias, bias_2 = (orig_bias, new_bias) if orig_bias is not None else (new_bias, None)
    txt, txt_2 = ('Original', 'New') if orig_bias is not None else ('New', '')
    
    colors = ['#34495e','#e67e22']
    
    attributes = [attr] if attr is not None else list(bias.keys())
    
    for attr in attributes:
        n_target = len(bias[attr])
        
        values = [target_value] if target_value is not None else list(bias[attr].keys())        
        for target_value in values:
            metrics = dict(bias[attr][target_value])
            height = int((len(metrics) + len(metrics)%2) / 2)
            fig, ax = plt.subplots(figsize=(15, 2*height))
            
            i = 1
            for metric in metrics:
                ax = plt.subplot(int(f'{height}2{i}'))
                value = metrics[metric]
                goal = utils.get_metric_goal(metric=metric)
                
                plots.plot_gauge(ax, value, goal, bar_color=colors[0])
                        
                if bias_2 is not None:
                    if attr in bias_2:
                        if target_value in bias_2[attr]:                  
                            if metric in bias_2[attr][target_value]:
                                value = bias_2[attr][target_value][metric]
                                ax.axvline(linewidth=4, color=colors[1], x=value)
                
                ax.set_title(f'{metric}')
                i += 1
                
            labels = [txt] if bias_2 is None else [txt, txt_2]
            fig.legend(labels=labels, fontsize=16)
                    
            fig.suptitle(f'Focus on {attr} attribute : bias plot for {target_value} label', fontsize=18)
            plots.plot_or_save(fname=f'monitoring_{attr}_{target_value}_gauge_bias_plot.png')


__all__ = [
    'BaseModel', 
    'ClassificationModel',
    'RegressionModel',
    'plot_classification_scores',
    'plot_confusion_matrix',
    'plot_roc_curve',
    'plot_class_distribution',
    'plot_curve',
    'compare_threshold_predictions'
    ]

from transparentai.models.base_model import BaseModel
from transparentai.models.classification_model import ClassificationModel
from transparentai.models.regression_model import RegressionModel
from transparentai.models.models_plots import *


__all__ = [
    'Monitoring', 
    'plot_bar_performance',
    'plot_gauge_bias'
    ]

from transparentai.monitoring.monitoring import Monitoring
from transparentai.monitoring.monitoring_plots import *


import pandas as pd
from IPython.display import display, Markdown

from sklearn.metrics import mean_absolute_error
from sklearn.metrics import mean_squared_error
from sklearn.metrics import r2_score

import transparentai.models.models_plots as plots
from transparentai.models.base_model import BaseModel


class RegressionModel(BaseModel):
    """
    Class to inspect a regression model based on a model which has a `predict`  function.
    It could help you to explore your model performance and validate or not your model.

    Example
    -------

    >>> from transparentai.models import ClassificationModel
    >>> model = RegressionModel(model=clf)
    >>> model.compute_scores(X=X_test, y=y_test)

    For more details please see the `RegressionModel with LinearRegression`_ notebook.

    .. _RegressionModel with LinearRegression : https://github.com/Nathanlauga/transparentai/notebooks/example_RegressionModel_regression.ipynb
    
    Attributes
    ----------
    X: pd.DataFrame
        X data that were used to get the predictions
    y_preds: np.array or pd.Series
        Predictions using X parameters
    y_true: np.array or pd.Series (optional)
        Real output
    model: 
        object with at least a `predict` function that returns
        a series or np.array
    scores_dict: dict
        Dictionary with metric name as keys and score value
        as values
    model_type: str
        String that indicates what model type it is : 'regression'
    """

    def __init__(self, model=None, X=None, y=None, y_preds=None):
        """
        Parameters
        ----------
        model:
            a classifier model that have a `predict` and 
            `predict_proba` (if classification) functions
        X: pd.DataFrame (optional)
            X data that will be used to get the predictions
        y: np.array or pd.Series (optional)
            Real output
        y_preds: np.array or pd.Series (optional)
            Predictions using X parameters if model is not set
        """
        super().__init__(model=model, X=X, y=y, y_preds=y_preds)
        self.model_type = 'regression'

        if (X is not None) & (y is not None) & (y_preds is not None):
            self.scores()

    def compute_scores(self, X=None, y=None):
        """
        Compute all predictions and scores

        Parameters
        ----------
        X: array-like of shape (n_samples, n_features)
            feature samples
        y: array-like of shape (n_samples,) or (n_samples, n_outputs)
            true labels for X.

        Raises
        ------
        AttributeError:
            X has to be set in init part or in the parameters
        AttributeError:
            y_true has to be set in init part or in the parameters as y
        AttributeError:
            model attribute was not set at the init step
        """
        if (X is None) & (self.X is None):
            raise AttributeError('X is mandatory to compute scores')
        if (y is None) & (self.y_true is None):
            raise AttributeError('y is mandatory to compute scores')
        if self.model is None:
            raise AttributeError('model attribute was not set at the init step')

        self.X = X if X is not None else self.X
        self.y_true = y if y is not None else self.y_true
        self.y_preds = self.model.predict(X)

        self.scores()

    def scores(self):
        """
        Compute classification metrics scores based on `skearn Model evaluation`_
        Current metrics : mean_absolute_error, mean_squared_error, root_mean_squared_error and r2_score

        .. _skearn Model evaluation: https://scikit-learn.org/0.15/modules/model_evaluation.html
        """
        scores = {}
        scores['MAE'] = mean_absolute_error(self.y_true, self.y_preds)
        scores['MSE'] = mean_squared_error(self.y_true, self.y_preds)
        scores['RMSE'] = mean_squared_error(self.y_true, self.y_preds, squared=True)
        scores['R2'] = r2_score(self.y_true, self.y_preds)

        self.scores_dict = scores

    
    def scores_to_json(self):
        """
        Formats scores_dict attributes to be read as a json object.

        If 'roc_auc' is in the keys then it computes the score mean
        for the different target values 

        Returns
        -------
        dict:
            Scores dict formated for json
        """
        scores_names = ['MAE', 'MSE', 'RMSE', 'R2']
        return self._scores_to_json(scores_names)

    def _plot_scores(self):
        """
        Display different charts for all the metrics.

        """
        # fun = plots.plot_classification_scores
        # self.plot_overall_scores(fun=fun, preds=self.y_proba)



import pandas as pd
import numpy as np

from transparentai.fairness.model_protected_attribute import ModelProtectedAttribute
from transparentai.fairness.bias_metric import BiasMetric


class ModelBiasMetric(BiasMetric):
    """
    Dataset class for bias metrics. It stored different metrics for a set of
    protected attributes (you can find some examples here_).

    List of bias metrics for DatasetBiasMetric : 

    1. Disparate impact
    2. Statistical parity difference
    3. Equal opportunity difference
    4. Average abs odds difference
    5. Theil index

    .. _here: https://www.fairwork.gov.au/employee-entitlements/protections-at-work/protection-from-discrimination-at-work

    Example
    -------
    For binary classification :

    .. code-block:: python

        from transparentai.fairness import ModelBiasMetric
        
        ...
        adult = ... # dataset
        clf = ... # train classifier
        preds = ... # predictions from classifier
        ...

        target = 'income'
        favorable_label = '>50K'
        dataset = StructuredDataset(df=adult, target=target)

        privileged_groups = {
            'marital-status': ['Married-civ-spouse','Married-AF-spouse'],
            'race': ['White'],
            'gender': ['Male']
        }   

        model_bias = ModelBiasMetric(dataset=dataset, preds=preds,
                                    privileged_groups=privileged_groups,
                                    favorable_label=favorable_label)

    Attributes
    ----------
    bias_type: str
        Bias type of the object : only 'dataset' or 'model'
    dataset: StructuredDataset
        dataset with a target set as attribute
    favorable_label:
        A specific value that is has an advantage over the other(s) 
        values
    labels: array like
        list of values inside target column
    protected_attributes: dict
        Dictionary with protected attributes as keys (e.g. gender) and
        ProtectedAttribute objects as values
    """

    def __init__(self, dataset, preds, privileged_groups, favorable_label=None):
        """
        Parameters
        ----------
        dataset: StructuredDataset
            dataset with a target set as attribute
        preds: array like (same length than dataset.labels)
            Predictions for the dataset 
        privileged_groups: dict(list())
            Dictionary with protected attributes as keys (e.g. gender) and
            list of value(s) that are considered privileged (e.g. ['Male'])
        privileged_values: dict
            Dictionnary with all protected attributes (category dtype) as key
            and a list of privileged value for the variable
            e.g. { 'gender' : ['Male'] }
        """
        super().__init__(dataset, privileged_groups, favorable_label)
        self.preds = preds
        protected_attributes = {}

        if (dataset.target_mean is not None):
            preds = np.where(preds > dataset.target_mean,
                             f'>{dataset.target_mean}', f'<={dataset.target_mean}')

        for attr in privileged_groups:
            values = privileged_groups[attr]
            protected_attr = ModelProtectedAttribute(dataset=dataset,
                                                     attr=attr,
                                                     privileged_values=values,
                                                     preds=preds,
                                                     favorable_label=favorable_label
                                                     )
            protected_attributes[attr] = protected_attr

        self.protected_attributes = protected_attributes
        self.set_bias_type(bias_type='model')



from IPython.display import display, Markdown
import json
import os

import transparentai.utils as utils

def print_md(text):
    """
    Display a text as Markdown.

    Parameters
    ----------
    text: str
        text to display

    Returns
    -------
    function:
        display(Markdown(text))
    """
    return display(Markdown(text))


def get_questions():
    """
    Retrieves questions to start.

    Questions are stored in a json file at ../src/json directory

    Returns
    -------
    dict:
        Dictionary with the questions classified in categories
    """
    fpath = os.path.abspath(os.path.join(
            os.path.dirname(__file__), '../src/json/'))

    with utils.OpenFile(fpath+'/questions.json', mode='r') as file:
        questions = json.load(file)
    file.close()
    return questions


def how_can_i_start(md=True):
    """
    Displays text about how to start an AI with this module.

    Parameters
    ----------
    md: bool (default True)
        Whether the text is display in Markdown or with print function
    """
    print_fun = print if not md else print_md

    title = 'How can I start' if not md else '### How can I start'
    print_fun(title)

    print_fun('This function is a helper to show some starting possibilities')
    print_fun('')

    print_fun('├── `transparentai.start.quick_start()` shows you questions about the project. ' +
              'If you complete it, at the end you will have answered questions about if your AI is viable.')
    print_fun('├── `transparentai.start.external_link()` shows you external references that can be more accurate ' +
              'to your AI.')


def quick_start(md=True, with_input=True):
    """
    Displays a set of questions to start a project. The questions were wrote using
    `ML Canvas`_ and `PAIR Guidebook`_.

    .. _ML Canvas: https://www.louisdorard.com/machine-learning-canvas
    .. _PAIR Guidebook: https://pair.withgoogle.com/

    Parameters
    ----------
    md: bool (default True)
        Whether the text is display in Markdown or with print function
    with_input: bool (default True)
        Whether you want to input answer or just see the questions 
    """
    print_fun = print if not md else print_md
    questions = get_questions()
    answered_question = questions # if answer is None else answer

    title = 'Quick start' if not md else '### Quick start'
    print_fun(title)

    canvas = 'ML Canvas' if not md else '[ML Canvas](https://www.louisdorard.com/machine-learning-canvas)'
    pair = 'PAIR Guidebook' if not md else '[PAIR Guidebook](https://pair.withgoogle.com/)'
    print_fun(f'Those questions were inspired by {canvas} and {pair}.')

    print_fun('')

    for section, sect_questions in questions.items():
        title = section if not md else f'#### {section}'
        print_fun(title)

        for k, v in sect_questions.items():
            q = v['question']
            print_fun(f'`{k}` : {q}')
            if with_input:
                ans = input('Answer here : ')
                answered_question[section][k]['answer'] = ans

    return answered_question


def format_answer(answer, format='json'):
    """
    Format the answer dictionnary to a single string or dictionnary
    depending on the format set.

    If format is 'json' then it returns a dictionary else a string.

    Parameters
    ----------
    answer: dict
        Dictionary with questions and answer 
        returned from `quick_start()` function
    format: str (default 'json')
        Format that you want for the output
        if json then a dictionary is returned else it's a string

    Returns
    -------
    dict or str
        Formated answers with questions
    """
    ans_return = '' if format != 'json' else {}

    questions = get_questions()

    for section, sect_questions in questions.items():
        for k, v in sect_questions.items():
            q = v['question']
            a = v['answer']
            if format == 'json':
                ans_return[k] = {
                    'question': q,
                    'answer': a
                }
            else:
                ans_return += f"{q}\nAnswer: {a}\n\n"

    return ans_return


def save_answer_to_file(answer, fname, format='json'):
    """
    Saves answer to a file.

    Parameters
    ----------
    answer: dict
        Dictionary with questions and answer 
        returned from `quick_start()` function
    fname: str
        string of the file path (including filename)
    format: str (default 'json')
        Format that you want for the output
        if json then a dictionary is stored else it's a string
    """
    answer = format_answer(answer, format=format)
    if format != 'json':
        utils.str_to_file(string=answer, fpath=fname)
    else:
        with utils.OpenFile(fname, mode='w') as file:
            json.dump(answer, file)
        file.close()


def format_link(url, md_text, md):
    """
    Formats url link depending on if it's Mardown or print function.

    Parameters
    ----------
    url: str
        Url string to the website
    md_text: str
        Text to display in hyperlink if it's markdown
    md: bool
        Whether the text is display in Markdown or with print function
    """
    return url if not md else f'[{md_text}]({url})'


def external_link(md=True):
    """
    Displays text about alternative solutions to start a ML project.

    Parameters
    ----------
    md: bool (default True)
        Whether the text is display in Markdown or with print function
    """
    print_fun = print if not md else print_md

    title = 'External links' if not md else '### External links'
    print_fun(title)
    print_fun(
        'This is the list of different links that can be usefull to start an AI (with more illustrations)')

    print_fun('')

    url = 'https://deon.drivendata.org/'
    md_text = 'deon'
    url = format_link(url, md_text, md=md)
    print_fun(f'├── {url} – Deon by DrivenData is a command line tool that allows you ' +
              'to easily add an ethics checklist to your data science projects.')

    url = 'https://www.louisdorard.com/machine-learning-canvas'
    md_text = 'ML Canvas'
    url = format_link(url, md_text, md=md)
    print_fun(f'├── {url} – it allows to design better Machine Learning systems. ' +
              'Keep teams of scientists, engineers and managers focused on the same objectives.')

    url = 'https://resources.sei.cmu.edu/library/asset-view.cfm?assetid=636620'
    md_text = 'Designing Ethical AI Experiences: Checklist and Agreement'
    url = format_link(url, md_text, md=md)
    print_fun(f'├── {url} – This document can be used to guide the development of accountable, ' +
              'de-risked, respectful, secure, honest, and usable artificial intelligence (AI) ' +
              'systems with a diverse team aligned on shared ethics.')

    url = 'https://www.ethicscanvas.org/index.html'
    md_text = 'Ethic Canvas Online'
    url = format_link(url, md_text, md=md)
    print_fun(f'├── {url} – A resource inspired by the traditional business canvas, which provides an ' +
              'interactive way to brainstorm potential risks, opportunities and solutions to ethical challenges ' +
              'that may be faced in a project using post-it note-like approach.')

    if md:
        print_fun('*****')
    else:
        print_fun('-----')

    url = 'https://github.com/EthicalML/awesome-artificial-intelligence-guidelines'
    detail = url if not md else f'[awesome-artificial-intelligence-guidelines]({url})'
    print_fun(f'You can find more tools and guideline here : {detail}')



import pandas as pd
import numpy as np
import os

import matplotlib.pyplot as plt
import seaborn as sns
from IPython.display import display, Markdown

from transparentai import plots


def plot_missing_values(df):
    """
    Show a bar plot that display percentage of missing values on columns that have some.
    If no missing value then it use `display` & `Markdown` functions to indicate it.

    Parameters
    ----------
    df: pd.DataFrame
        Dataframe to inspect
    """
    df_null = pd.DataFrame(len(df) - df.notnull().sum(), columns=['Count'])
    df_null = df_null[df_null['Count'] > 0].sort_values(
        by='Count', ascending=False)
    df_null = df_null/len(df)*100

    if len(df_null) == 0:
        display(Markdown('No missing value.'))
        return

    x = df_null.index.values
    height = [e[0] for e in df_null.values]

    fig, ax = plt.subplots(figsize=(20, 5))
    ax.bar(x, height, width=0.8)
    plt.xticks(x, x, rotation=30)
    plt.xlabel('Columns')
    plt.ylabel('Percentage')
    plt.title('Percentage of missing values in columns')

    plots.plot_or_save(fname='missing_values.png')


def plot_numerical_var(df, var, target=None):
    """
    Show variable information in graphics for numerical variables.
    At least the displot & boxplot.
    If target is set 2 more plots : stack plot and stack plot with percentage

    Parameters
    ----------
    df: pd.DataFrame
        Dataframe to inspect
    var: str
        Column name that contains numerical values
    target: str (default None)
        Target column for classifier
    """
    rows = 1 if target is None else 2
    fig, ax = plt.subplots(figsize=(16, rows*5))

    ax = plt.subplot(int(f'{rows}21'))
    if target == None:
        sns.distplot(df[var].dropna())
    else:
        labels = sorted(df[target].unique())
        for l in labels:
            df_target = df[df[target] == l]
            if df_target[var].nunique() <= 1:
                sns.distplot(df_target[var], kde=False)
            else:
                sns.distplot(df_target[var])
            del df_target

    ax = plt.subplot(int(f'{rows}22'))
    x = df[target] if target != None else None

    sns.boxplot(x=x, y=df[var].dropna())

    if target != None:
        tab = pd.crosstab(df[var], df[target])

        ax = plt.subplot(223)
        plots.plot_stack(ax=ax, df=tab, labels=labels)

        tab = tab.div(tab.sum(axis=1), axis=0)

        ax = plt.subplot(224)
        plots.plot_stack(ax=ax, df=tab, labels=labels)

    plots.plot_or_save(fname=f'{var}_variable_plot.png')


def plot_categorical_var(df, var, target=None):
    """
    Show variable information in graphics for categorical variables.
    For 10 most frequents values : bar plot and a pie chart

    If target is set : plot stack bar for 10 most frequents values

    Parameters
    ----------
    df: pd.DataFrame
        Dataframe to inspect
    var: str
        Column name that contains categorical values
    target: str (default None)
        Target column for classifier
    """
    val_cnt = df[var].value_counts()
    if len(val_cnt) > 10:
        val_cnt = val_cnt.head(10)

    labels = val_cnt.index
    sizes = val_cnt.values
    colors = sns.color_palette("Blues", len(labels))

    rows = 1 if target is None else 2
    fig, ax = plt.subplots(figsize=(16, rows*5))

    ax = plt.subplot(int(f'{rows}21'))
    ax.bar(labels, sizes, width=0.8)
    plt.xticks(labels, labels, rotation=60)

    ax = plt.subplot(int(f'{rows}22'))
    ax.pie(sizes, labels=labels, colors=colors,
           autopct='%1.0f%%', shadow=True, startangle=130)
    ax.axis('equal')
    ax.legend(loc=0, frameon=True)

    if target != None:
        ax = plt.subplot(212)

        legend_labels = sorted(df[target].unique())
        tab = pd.crosstab(df[var], df[target])
        tab = tab.loc[labels]
        plots.plot_stack_bar(ax=ax, df=tab, labels=labels,
                             legend_labels=legend_labels)

    plots.plot_or_save(fname=f'{var}_variable_plot.png')


def plot_datetime_var(df, var, target=None):
    """
    Show variable information in graphics for datetime variables.
    Display only the time series line if no target is set else, it shows
    2 graphics one with differents lines by value of target and one stack line plot

    If difference between maximum date and minimum date is above 1000 then plot by year.

    Parameters
    ----------
    df: pd.DataFrame
        Dataframe to inspect
    var: str
        Column name that contains datetime values
    target: str (default None)
        Target column for classifier
    """
    df = df.copy()
    rows = 1 if target is None else 2
    fig, ax = plt.subplots(figsize=(16, rows*5))

    date_min = df[var].min()
    date_max = df[var].max()
    if (date_max - date_min).days > 1000:
        df[var] = df[var].dt.year

    if target == None:
        val_cnt = df[var].value_counts()
        sns.lineplot(data=val_cnt)
    else:
        ax = plt.subplot(121)

        legend_labels = sorted(df[target].unique())
        tab = pd.crosstab(df[var], df[target])
        sns.lineplot(data=tab)

        tab.div(tab.sum(axis=1), axis=0)

        ax = plt.subplot(122)
        plots.plot_stack(ax=ax, df=tab, labels=legend_labels)

    plots.plot_or_save(fname=f'{var}_variable_plot.png')


def display_meta_var(df, var):
    """
    Display some meta informations about a specific variable of a given dataframe
    Meta informations : # of null values, # of uniques values and 2 most frequent values

    Parameters
    ----------
    df: pd.DataFrame
        Dataframe to inspect
    var: str
        Column name that is in df        
    """
    nb_null = df[var].isnull().sum()
    nb_uniq = df[var].nunique()
    val_cnt = df[var].value_counts()
    n = 2 if len(val_cnt) > 1 else len(val_cnt)
    most_freq = df[var].value_counts().head(n).to_dict()
    display(Markdown(
        f'**{var} :** {nb_null} nulls, {nb_uniq} unique vals, most common: {most_freq}'))


def plot_numerical_jointplot(df, var1, var2, target=None):
    """
    Show two numerical variables relations with jointplot.

    Parameters
    ----------
    df: pd.DataFrame
        Dataframe to inspect
    var1: str
        Column name that contains first numerical values
    var2: str
        Column name that contains second numerical values
    target: str (default None)
        Target column for classifier
    """
    if df[var1].nunique() <= 1:
        display(Markdown(f'{var1} has only one value'))
        return
    if df[var2].nunique() <= 1:
        display(Markdown(f'{var2} has only one value'))
        return

    if target is not None:
        if (min(df[target].value_counts()) <= 1) | (df[target].nunique() == 1):
            display(
                Markdown(f'{target} has not enough data to be represented'))
            target = None

    if target == None:
        g = sns.jointplot(var1, var2, data=df, space=0, height=8)
    else:
        legend_labels = sorted(df[target].unique())
        cols = [var1, var2, target] if target is not None else [var1, var2]
        df = df[cols].dropna()
        grid = sns.JointGrid(x=var1, y=var2, data=df, height=7)

        g = grid.plot_joint(sns.scatterplot, hue=target, data=df, alpha=0.3)
        for l in legend_labels:
            df_v1 = df.loc[df[target] == l, var1]
            df_v2 = df.loc[df[target] == l, var2]
            sns.distplot(df_v1, ax=g.ax_marg_x, kde=(df_v1.nunique() > 1))
            sns.distplot(df_v2, ax=g.ax_marg_y, kde=(
                df_v2.nunique() > 1), vertical=True)

    plots.plot_or_save(fname=f'{var1}_{var2}_variable_jointplot.png')


def plot_one_cat_and_num_variables(df, var1, var2, target=None):
    """
    Show boxplots for a specific pair of categorical and numerical variables
    If target is set, separate dataset for each target value.

    Parameters
    ----------
    df: pd.DataFrame
        Dataframe to inspect
    var1: str
        Column name that contains categorical values
    var2: str
        Column name that contains numerical values
    target: str (default None)
        Target column for classifier
    """
    if df[var1].nunique() <= 1:
        display(Markdown(f'{var1} has only one value'))
        return
    if df[var2].nunique() <= 1:
        display(Markdown(f'{var2} has only one value'))
        return

    val_cnt = df[var1].value_counts()
    if len(val_cnt) > 10:
        val_cnt = val_cnt.head(10)

    df_plot = df[df[var1].apply(lambda x: x in val_cnt.index.values)]

    fig, ax = plt.subplots(figsize=(16, 5))
    palette = "Blues" if target == None else "colorblind"
    sns.boxplot(x=var1, y=var2, hue=target,
                data=df_plot, palette=palette)
    plt.xticks(rotation=40)

    plots.plot_or_save(fname=f'{var1}_{var2}_variable_boxplot.png')


def plot_correlation_matrix(corr_df, fname='correlation_matrix_plot.png'):
    """
    Plot a seaborn heatmap based on a correlation dataframe.

    Parameters
    ----------
    corr_df: pd.DataFrame
        Correlation dataframe
    fname: str
        File path where to save the plot if plots.save_plot if True
    """
    annot = max(corr_df.shape) <= 20

    fig, ax = plt.subplots(figsize=(11, 12))
    ax = sns.heatmap(
        corr_df,
        vmin=-1, vmax=1, center=0,
        cmap=sns.color_palette("RdBu_r", 100),
        square=(corr_df.shape[0] == corr_df.shape[1]),
        annot=annot,
        fmt='.2f'
    )
    ax.set_xticklabels(
        ax.get_xticklabels(),
        rotation=45,
        horizontalalignment='right'
    )
    plots.plot_or_save(fname=fname)



__all__ = [
    'BiasMetric', 
    'DatasetBiasMetric', 
    'ModelBiasMetric', 
    'ProtectedAttribute', 
    'DatasetProtectedAttribute',
    'ModelProtectedAttribute',
    'plot_bias_metrics',
    'plot_dataset_bias_metrics',
    'plot_model_bias_metrics',
    'plot_protected_attr_row_model',
    'plot_bar_freq_predicted',
    'plot_protected_attr_row',
    'plot_pie',
    'plot_percentage_bar_man_img'
    ]
    
from transparentai.fairness.bias_metric import BiasMetric
from transparentai.fairness.dataset_bias_metric import DatasetBiasMetric
from transparentai.fairness.model_bias_metric import ModelBiasMetric
from transparentai.fairness.protected_attribute import ProtectedAttribute
from transparentai.fairness.dataset_protected_attribute import DatasetProtectedAttribute
from transparentai.fairness.model_protected_attribute import ModelProtectedAttribute
from transparentai.fairness.fairness_plots import *


import numpy as np
import pandas as pd
import warnings

from transparentai.models import ClassificationModel, RegressionModel
from transparentai.datasets import StructuredDataset
from transparentai.fairness import DatasetBiasMetric, ModelBiasMetric
import transparentai.monitoring.monitoring_plots as plots


class Monitoring():
    """
    This class allows to see if your model is still with the same performance than
    just after training and gives you some insight with graphics about the current performance 
    and dataset / model bias.  

    Example
    -------
    >>> from transparentai.monitoring import Monitoring
    >>> monitoring = Monitoring(X=new_X, y_preds=new_y_preds, 
                                y_real=new_y_real,
                                privileged_groups=privileged_groups,
                                alert_threshold=alert_threshold,
                                model_type='classification')
 
    For more details please see the `Monitoring for binary classification`_ or
    `Montioring for regression`_ notebooks.

    .. _Monitoring for binary classification : https://github.com/Nathanlauga/transparentai/notebooks/example_monitoring_binary_classification.ipynb
    .. _Montioring for regression : https://github.com/Nathanlauga/transparentai/notebooks/example_monitoring_regression.ipynb

    Attributes
    ----------
    X: pd.DataFrame
        New X data that were used to get the predictions
    y_preds: np.array or pd.Series
        Predictions using X parameters
    y_real: np.array or pd.Series (optional)
        Real output
    model_type: str
        'classification' or 'regression'
    orig_metrics: dict(dict)
        Dictionary with metrics group in keys ('performance', 'bias_dataset', 'bias_model') 
        and original metric values in values
    new_metrics: dict(dict)
        Dictionary with metrics group in keys ('performance', 'bias_dataset', 'bias_model') 
        and new metric values in values
    privileged_groups: dict(list())
        Dictionary with protected attributes as keys (e.g. gender) and
        list of value(s) that are considered privileged (e.g. ['Male'])
    alert_threshold: dict (optional)
        Dictionary with metrics in keys and
        threshold values in values
    dateset: transparentai.datasets.StructuredDataset
        Dataset with X and either y_preds if y_real is None or 
        y_real if not.
    """

    def __init__(self, X, y_preds, y_real=None, model_type='classification',
                 orig_metrics=None, privileged_groups=None, alert_threshold=None, mean=None):
        """
        Parameters
        ----------
        X: pd.DataFrame
            New X data that were used to get the predictions
        y_preds: np.array or pd.Series
            Predictions using X parameters
        y_real: np.array or pd.Series (optional)
            Real output
        model_type: str
            'classification' or 'regression'
        orig_metrics: dict(dict)
            Dictionary with metrics group in keys ('performance', 'bias_dataset', 'bias_model') 
            and original metric values in values
        privileged_groups: dict(list())
            Dictionary with protected attributes as keys (e.g. gender) and
            list of value(s) that are considered privileged (e.g. ['Male'])
        alert_threshold: dict (optional)
            Dictionary with metrics in keys and
            threshold values in values
        mean: number
            Mean of target (for regression) to get the same label than
            original metrics for bias.

        Raises
        ------
        ValueError
            Only regression and classification are handled  for model_type.
        TypeError
            X has to be a pandas dataframe
        ValueError
            y_preds and X must have the same length
        ValueError
            y_real and X must have the same length
        ValueError
            Valid keys for orig_metrics are 'performance', 'bias_dataset' or 'bias_model'.
        """
        if model_type not in ['classification', 'regression']:
            raise ValueError(
                'Only regression and classification are handled  for model_type.')
        if type(X) not in [pd.DataFrame]:
            raise TypeError('X has to be a pandas dataframe')
        if len(X) != len(y_preds):
            raise ValueError('y_preds and X must have the same length')
        if y_real is not None:
            if len(X) != len(y_real):
                raise ValueError('y_real and X must have the same length')
        if orig_metrics is not None:
            metrics_keys = ['performance', 'bias_dataset', 'bias_model']
            if not any([k in metrics_keys for k in orig_metrics]):
                raise ValueError(
                    "Valid keys for orig_metrics are 'performance', 'bias_dataset' or 'bias_model'.")

        self.X = X
        self.y_preds = y_preds
        self.y_real = y_real
        self.model_type = model_type
        self.orig_metrics = orig_metrics
        self.privileged_groups = privileged_groups
        self.alert_threshold = alert_threshold

        df = X.copy().reset_index(drop=True)
        if type(y_preds) in [pd.Series, pd.DataFrame]:
            y_preds = y_preds.values
        if y_real is not None:
            if type(y_real) in [pd.Series, pd.DataFrame]:
                y_real = y_real.values

        df['target'] = y_preds if y_real is None else y_real

        target_regr = model_type == 'regression'
        self.dataset = StructuredDataset(df=df, target='target', mean=mean, target_regr=target_regr)

        self._compute_new_metrics()

        if orig_metrics is not None:
            self._check_orig_and_new_metrics()

    def _compute_orig_metrics(self, X_orig, y_orig):
        """
        Only if you don't have original metrics already stored
        """
        # Todo
        orig_metrics = {}

        self.orig_metrics = orig_metrics

    def _compute_new_metrics(self):
        """
        Computes new metrics based on given attributes in init.

        If y_real is not None then computes new performance scores.

        If privileged_groups then it also computes dataset bias and if y_real
        is define the model bias
        """
        new_metrics = {}
        # handle only 2 first ?
        # model_bias_metrics = ['Disparate impact', 'Statistical parity difference']
        # model_bias_metrics += ['Equal opportunity difference', 'Average abs odds difference', 'Theil index']

        # I have y_real ==> compute model perf & define model bias to compute
        if (self.y_real is not None):
            if self.model_type == 'classification':
                model_obj = ClassificationModel
            elif self.model_type == 'regression':
                model_obj = RegressionModel

            model = model_obj(X=self.X, y=self.y_real, y_preds=self.y_preds)
            new_metrics['performance'] = model.scores_to_json()

        # I have protected attr ==> compute dataset bias & model bias
        if (self.privileged_groups is not None):

            bias = DatasetBiasMetric(self.dataset, self.privileged_groups)
            new_metrics['bias_dataset'] = bias.metrics_to_json()

            if self.y_real is not None:
                bias = ModelBiasMetric(
                    self.dataset, self.y_preds, self.privileged_groups)
                new_metrics['bias_model'] = bias.metrics_to_json()

        self.new_metrics = new_metrics

    def _check_one_metric(self, metric):
        """
        Compares differents keys of a specified metric.

        Show a warning if a key is in a dictionary but not in the other one.

        Parameters
        ----------
        metric: str
            Metric key to compare in both orig_metrics and
            new_metrics
        """
        if (metric in self.orig_metrics) & (metric in self.new_metrics):
            not_in_new = [k for k in self.orig_metrics[metric]
                          if k not in self.new_metrics[metric]]
            not_in_orig = [k for k in self.new_metrics[metric]
                           if k not in self.orig_metrics[metric]]
            for k in not_in_new:
                warnings.warn(
                    f"In original {metric} dict '{k}' key is in but not in the new one.")
            for k in not_in_orig:
                warnings.warn(f"In new {metric} dict '{k}' key is in but not in the orignal one.")

    def _check_orig_and_new_metrics(self):
        """
        Checks that performance metrics in both new and original dictionaries
        are the same.
        """
        if (self.orig_metrics is None) | (self.new_metrics is None):
            return
        # Check performance
        self._check_one_metric(metric='performance')

    def plot_perfomance(self):
        """
        Plot performance with one bar chart for each metric.

        If `orig_metrics` is not None but `new_metrics` is, then plots only original 
        performance metrics.
        If `orig_metrics` is None but `new_metrics` is not, then plots only new 
        performance metrics.
        If neither `orig_metrics` and `new_metrics` are not None, then plots
        comparison between original metrics and new ones.
        """
        if (self.orig_metrics is None) & (self.y_real is None):
            raise ValueError(
                "Either original metrics or y_real has to bas set in init to show performance")
        if self.orig_metrics is not None:
            if ('performance' not in self.orig_metrics) & (self.y_real is None):
                raise ValueError(
                    "'performance' has to be in orig_metrics dict")

        if 'performance' not in self.new_metrics:
            orig_perf = self.orig_metrics['performance']
            plots.plot_bar_performance(orig_perf=orig_perf,
                                       alert_threshold=self.alert_threshold)

        elif self.orig_metrics is None:
            new_perf = self.new_metrics['performance']
            plots.plot_bar_performance(
                new_perf=new_perf, alert_threshold=self.alert_threshold)

        else:
            orig_perf = self.orig_metrics['performance']
            new_perf = self.new_metrics['performance']
            plots.plot_bar_performance(
                orig_perf, new_perf, self.alert_threshold)

    def _plot_bias(self, bias_key='bias_dataset', attr=None, target_value=None):
        """
        Plot bias with one gauge plot for each metric.
        
        If `attr` and `target_value` it reduces the number of plots 
        Example : if you have 2 target values (e.g. 0 and 1) and different protected
        attributes (e.g. gender and marital status) then you can say that it will
        only plots for gender and target value = 1 

        If `orig_metrics` is not None but `new_metrics` is, then plots only original 
        performance metrics.
        If `orig_metrics` is None but `new_metrics` is not, then plots only new 
        performance metrics.
        If neither `orig_metrics` and `new_metrics` are not None, then plots
        comparison between original metrics and new ones.

        Parameters
        ----------
        bias_key: str (default 'bias_dataset')
            Key for bias metrics
            Can be 'bias_dataset' or 'bias_model'
        attr: str (optional)
            Name of the attribute to analyse
        target_value: (optional)
            Specific label value 
        """
        if (self.orig_metrics is None) & (self.y_real is None):
            raise ValueError(
                "Either original metrics or y_real has to bas set in init to show bias")
        if self.orig_metrics is not None:
            if self.privileged_groups is None:
                raise ValueError(
                    f"privileged_groups object attribute was not set at the init")

            if (bias_key not in self.orig_metrics) & (self.y_real is None):
                raise ValueError(
                    f"'{bias_key}' has to be in orig_metrics dict when y_real is not set")

        if bias_key not in self.new_metrics:
            orig_bias = self.orig_metrics[bias_key]
            plots.plot_gauge_bias(attr, target_value, orig_bias=orig_bias)

        elif self.orig_metrics is None:
            new_bias = self.new_metrics[bias_key]
            plots.plot_gauge_bias(attr, target_value, new_bias=new_bias)

        else:
            orig_bias = self.orig_metrics[bias_key]
            new_bias = self.new_metrics[bias_key]
            plots.plot_gauge_bias(attr, target_value, orig_bias, new_bias)

    def plot_dataset_bias(self, attr=None, target_value=None):
        """
        Plot bias with one gauge plot for each dataset bias metric.

        Parameters
        ----------
        attr: str (optional)
            Name of the attribute to analyse
        target_value: (optional)
            Specific label value 
        """
        self._plot_bias(bias_key='bias_dataset', attr=attr,
                        target_value=target_value)

    def plot_model_bias(self, attr=None, target_value=None):
        """
        Plot bias with one gauge plot for each model bias metric.

        Parameters
        ----------
        attr: str (optional)
            Name of the attribute to analyse
        target_value: (optional)
            Specific label value 
        """
        self._plot_bias(bias_key='bias_model', attr=attr,
                        target_value=target_value)



import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches

from transparentai import plots

def plot_global_feature_influence(feat_importance):
    """
    Display global feature influence sorted.
    
    Parameters
    ----------
    feat_importance: pd.Series
        Feature importance with feature as indexes and 
        shap value as values
    """
    fig, ax = plt.subplots(figsize=(12, 8))
    ax.barh(feat_importance.index, feat_importance.values, label=feat_importance.index, align='center')
    plt.title('Feature importance (using Shap)')
    plt.ylabel('Features')
    plt.xlabel('Global value influence')
    
    plots.plot_or_save(fname='global_feature_influence_plot.png')
           
        
def plot_local_feature_influence(feat_importance, base_value, pred):
    """
    Display local feature influence sorted for a specific
    prediction.
    
    Parameters
    ----------
    feat_importance: pd.Series
        Feature importance with feature as indexes and 
        shap value as values
    base_value: number
        prediction value if we don't put any feature into the model
    pred: number
        predicted value
    """    
    feat_importance = feat_importance.sort_values(ascending=False)
    
    current_val = base_value
    left = list()
    for feat, value in feat_importance.items():
        left.append(current_val)
        current_val += value
    left = list(reversed(left))
        
    feat_importance = feat_importance.sort_values()
        
    fig, ax = plt.subplots(figsize=(12, 8))
    line_based = ax.axvline(x=base_value, linewidth=2, color='black', linestyle='--')
    line_pred = ax.axvline(x=pred, linewidth=2, color='black')
    
    colors = ['green' if v > 0 else 'red' for v in feat_importance.values]
    
    rects = ax.barh(feat_importance.index, 
                    feat_importance.values, 
                    label=feat_importance.index, 
                    color=colors,
                    left=left,
                    align='center')    
    
    for r in rects:
        h, w, x, y = r.get_height(), r.get_width(), r.get_x(), r.get_y()
        if h == 0:
            continue
        val = round(w,4)
        x = x+w if val > 0 else x
        ax.annotate(str(val), xy=(x, y+h/2), xytext=(0, 0),
                    textcoords="offset points", va='center',
                    color='black', clip_on=True)
    
    xlim0, xlim1 = ax.get_xlim()
    size = xlim1-xlim0
    ax.set_xlim(xlim0-size*0.001, xlim1+size*0.1)
    plt.title('Feature importance (using Shap)')
    plt.ylabel('Features')
    plt.xlabel('Local value influence')
    
    green_patch = mpatches.Patch(color='green')
    ax.legend([line_based, line_pred, green_patch, rects], 
              ['Based value', 'Prediction', 'Positive influence', 'Negative influence'])
    
    plots.plot_or_save(fname='local_feature_influence_plot.png')



import os
import pandas as pd
import numpy as np
import scipy.stats as ss
from sklearn.preprocessing import LabelEncoder
import json


class OpenFile:
    """
    Class that open a file and close it at the end
    Attributes
    ----------
    fname : str
        file name
    mode : str
        mode for open() method
    """

    def __init__(self, fname: str, mode='r'):
        self.fname = fname
        self.mode = mode

    def __enter__(self):
        self.file = open(self.fname, self.mode)
        return self.file

    def __exit__(self, type, value, traceback):
        self.file.close()


def str_to_file(string: str, fpath: str):
    """
    Create a file based on a string.

    Parameters
    ----------
    string: str
        string to write into the file
    fname: str
        file name
    """
    with OpenFile(fpath, 'w') as file:
        file.write(string)
    file.close()


def remove_var_with_one_value(df):
    """
    Removes dataset's columns that only contain one unique value.

    Parameters
    ----------
    df: pd.DataFrame
        Dataframe to inspect

    Returns
    -------
    pd.DataFrame:
        Dataframe without columns with only one unique value
    """
    if len(df) <= 1:
        return df

    drop_cols = list()
    for var in df:
        if df[var].nunique() <= 1:
            drop_cols.append(var)
    return df.drop(columns=drop_cols)


def encode_categorical_vars(df):
    """
    Encode categorical variables from a dataframe to be numerical (discrete)
    It uses LabelEncoder classes from scikit-learn

    Parameters
    ----------
    df: pd.DataFrame
        Dataframe to update

    Returns
    -------
    pd.DataFrame:
        Encoded dataframe
    dict:
        Encoders with feature name on keys and
        encoder as value
    """
    cat_vars = df.select_dtypes(['object', 'category']).columns
    data_encoded = df.copy()
    for var in df.select_dtypes('category').columns:
        data_encoded[var] = data_encoded[var].cat.add_categories('Unknown')

    data_encoded[cat_vars] = data_encoded[cat_vars].fillna('Unknown')

    # Use Label Encoder for categorical columns (including target column)
    encoders = {}
    for feature in cat_vars:
        le = LabelEncoder()
        le.fit(data_encoded[feature].dropna())

        data_encoded[feature] = le.transform(data_encoded[feature])
        encoders[feature] = le

    return data_encoded, encoders


def cramers_v(x, y):
    """
    Returns the Cramer V value of two categorical variables using
    chi square. This correlation metric is between 0 and 1.

    Code source found in this article : 
    https://towardsdatascience.com/the-search-for-categorical-correlation-a1cf7f1888c9

    Parameters
    ----------
    x: array like
        first categorical variable
    y: array like
        second categorical variable

    Returns
    -------
    float:
        Cramer V value
    """
    confusion_matrix = pd.crosstab(x, y)
    chi2 = ss.chi2_contingency(confusion_matrix)[0]
    n = confusion_matrix.sum().sum()
    phi2 = chi2/n
    r, k = confusion_matrix.shape
    phi2corr = max(0, phi2-((k-1)*(r-1))/(n-1))
    rcorr = r-((r-1)**2)/(n-1)
    kcorr = k-((k-1)**2)/(n-1)
    if min((kcorr-1), (rcorr-1)) == 0:
        return 0
    return np.sqrt(phi2corr/min((kcorr-1), (rcorr-1)))


def init_corr_matrix(columns, index, fill_diag=1.):
    """
    Return a matrix n by m fill of 0 (except on the diagonal if squared matrix)
    Recommended for correlation matrix

    Parameters
    ----------
    columns: 
        list of column names
    index:
        list of index names
    fill_diag: float
        if squared matrix, then set diagonal with this value

    Returns
    -------
    pd.DataFrame
        Initialized matrix
    """
    zeros = np.zeros((len(index), len(columns)), float)
    if len(columns) == len(index):
        rng = np.arange(len(zeros))
        zeros[rng, rng] = fill_diag
    return pd.DataFrame(zeros, columns=columns, index=index)


def regression_to_classification(df, target, mean):
    """
    Convert a dataframe for regression to classification by 
    computing if the value is above the average.

    Parameters
    ----------
    df: pd.DataFrame
        Dataframe to update
    target: str
        Target column that has to be a numerical column

    Returns
    -------
    pd.DataFrame
        Dataframe updated with target column as a category variable
    np.array
        Original values of the target column

    Raises
    ------
    ValueError:
        `target` parameter has to be a numeric variable
    """
    if target not in df.select_dtypes('number').columns:
        raise ValueError('target column is not a number')

    orig_target_val = df[target].values
    df[target] = np.where(orig_target_val > mean, f'>{mean}', f'<={mean}')
    df[target] = df[target].astype('category')

    return df, orig_target_val


def labelencoder_to_dict(encoder):
    """
    Convert a LabelEncoder class from scikit-learn 
    to a dictionary with index as key and original value as value

    Example:
    encoder.classes\_ is ['Male', 'Female']
    returns {0:'Male', 1:'Female'}

    Parameters
    ----------
    encoder: LabelEncoder
        encoder fitted

    Returns
    -------
    dict:
        encoder transformed

    Raises
    ------
    TypeError:
        `encoder` has to be a `LabelEncoder` 
        from `sklearn.preprocessing` module
    """
    if type(encoder) != LabelEncoder:
        raise TypeError('encoder has to be a LabelEncoder.')

    feat_classes = encoder.classes_
    classes_dict = {}
    for i, val in enumerate(feat_classes):
        classes_dict[i] = val

    return classes_dict


def get_metric_goal(metric):
    """
    Returns bias metric goal given metric name.

    Parameters
    ---------- 
    metric: str
        Metric's name 

    Returns
    -------
    number
        Bias metric's goal
    """
    if metric == 'Disparate impact':
        return 1
    else:
        return 0

def save_dict_to_json(obj, fname):
    """
    Save a dictionary object to json file.

    Parameters
    ----------
    obj: dict
        dictionary to save
    fname: str
        string of the file path (including filename)
    """
    with OpenFile(fname, mode='w') as file:
        json.dump(obj, file)
    file.close()

def reduce_df_nrows(df, nrows):
    """
    Returns a dataframe reduced of n rows using `sample()` method.

    Parameters
    ----------
    df: pd.DataFrame
        Dataframe to reduce
    nrows: int
        Number of rows to keep
    """
    if type(df) != pd.DataFrame:
        raise TypeError('df has to be a pandas DataFrame object')

    if len(df) <= nrows:
        return df

    df_filter = df.sample(nrows)
    return df_filter


import pandas as pd
import numpy as np
from IPython.display import display, Markdown
from abc import abstractmethod

from transparentai.datasets.structured_dataset import StructuredDataset
import transparentai.fairness.fairness_plots as plots
import transparentai.utils as utils


class BiasMetric():
    """
    Base class for bias metrics. It stored different metrics for a set of
    protected attributes (you can find some examples here_).

    .. _here: https://www.fairwork.gov.au/employee-entitlements/protections-at-work/protection-from-discrimination-at-work

    Attributes
    ----------
    bias_type: str
        Bias type of the object : only 'dataset' or 'model'
    dataset: StructuredDataset
        dataset with a target set as attribute
    favorable_label:
        A specific value that is has an advantage over the other(s) values
    labels: array like
        list of values inside target column
    protected_attributes: dict
        Dictionary with protected attributes as keys (e.g. gender) and
        ProtectedAttribute objects as values
    """
    bias_type = None

    def __init__(self, dataset, privileged_groups, favorable_label=None):
        """
        Parameters
        ----------
        dataset: StructuredDataset
            dataset with a target set as attribute
        privileged_groups: dict(list())
            Dictionary with protected attributes as keys (e.g. gender) and
            list of value(s) that are considered privileged (e.g. ['Male'])
        privileged_values: dict
            Dictionnary with all protected attributes (category dtype) as key
            and a list of privileged value for the variable
            e.g. { 'gender' : ['Male'] }
        """

        if not isinstance(dataset, StructuredDataset):
            raise TypeError(
                'dataset attribute must be a StructuredDataset object!')
        if dataset.target is None:
            raise ValueError('target from StructuredDataset has to be set')
        if (favorable_label is not None) and (favorable_label not in dataset.df[dataset.target].unique()):
            raise ValueError(
                'favorable_label has to be a value of dataset target columns.')
        if privileged_groups is None:
            raise ValueError('privileged_groups has to be set')
        if (privileged_groups is not None) and (any([v not in dataset.df.columns for v in privileged_groups])):
            raise ValueError('privileged variables not in dataset')

        self.dataset = dataset
        self.favorable_label = favorable_label
        self.labels = dataset.df[dataset.target]

    def get_bias_metrics(self, attr=None):
        """
        Retrieve all bias metrics of protected attributes and format
        them to a DataFrame.

        Parameters
        ----------
        attr: str (default None)
            Protected attribute to inspect (if None display bias for all attributes)

        Returns
        -------
        pd.DataFrame
            formated bias metrics dataframe
        """
        if attr is None:
            metrics = list()
            for attr in self.protected_attributes:
                attr_metrics = self.protected_attributes[attr].metrics
                attr_metrics['attr'] = attr
                attr_metrics = attr_metrics.reset_index().set_index([
                    'attr', 'index'])
                metrics.append(attr_metrics)
            metrics = pd.concat(metrics)
            # metrics.index = list(self.protected_attributes.keys())
            # metrics.columns = ['Metrics dataframe']
        else:
            metrics = self.protected_attributes[attr].metrics
            if 'attr' in metrics.columns:
                metrics = metrics.drop(columns='attr')
        return metrics

    def _metrics_one_attr_to_json(self, attr, target_value=None):
        """
        Returns a dictionary formated containing bias metrics for a
        specific attribute.

        Parameters
        ----------
        attr: str
            Protected attribute to inspect
        target_value: str (default None)
            Specific value of the target (if None display bias for all target values)

        Returns
        -------
        dict(dict):
            Dictionary with target value(s) as keys and dictionary with metrics 
            name and values as values.
        """
        metrics = self.get_bias_metrics(attr=attr)

        if (target_value is None) and (self.favorable_label is not None):
            target_value = self.favorable_label

        attr_json = {}
        if target_value is None:
            for value in self.labels.unique():
                attr_json[value] = metrics.loc[value].to_dict()
        else:
            attr_json[target_value] = metrics.loc[target_value].to_dict()

        return attr_json

    def metrics_to_json(self, attr=None, target_value=None):
        """
        Returns a dictionary formated containing bias metrics.

        Parameters
        ----------
        attr: str (default None)
            Protected attribute to inspect (if None returns bias for all attributes)
        target_value: str (default None)
            Specific value of the target (if None display bias for all target values)
            It's usefull to avoid repetition in case of a binary classifier

        Returns
        -------
        dict(dict):
            Dictionary with attribute as keys and dictionary with metrics 
            (returned by _metrics_one_attr_to_json() function) as values.
        """
        if attr is None:
            metrics_json = {}
            for attr in self.protected_attributes:
                metrics_json[attr] = self._metrics_one_attr_to_json(
                    attr=attr, target_value=target_value)
        else:
            metrics_json = {attr: self._metrics_one_attr_to_json(
                attr=attr, target_value=target_value)}

        return metrics_json

    def save_bias_metrics(self, fname):
        """
        Save metrics to a json file.

        Parameters
        ----------
        fname: str
            string of the file path (including filename)
        """
        metrics_json = self.metrics_to_json()
        utils.save_dict_to_json(obj=metrics_json, fname=fname)

    def __metrics_is_biased(self, attr=None):
        """
        Returns a Dataframe with booleans as values that indicates if 
        a metric is considered biased or not.

        Parameters
        ----------
        attr: str (default None)
            Protected attribute to inspect (if None returns bias for all attributes)

        Returns
        -------
        pd.DataFrame
            Dataframe with the same size than get_bias_metrics() return
            with boolean as values
        """
        metrics = self.get_bias_metrics(attr=attr)
        copy = metrics.copy()
        for metric in metrics:
            copy[metric] = utils.get_metric_goal(metric)

        return (metrics - copy).applymap(lambda x: (x > 0.2) or (x < -0.2))

    def __insight_str(self, metrics_biased, attr, target_value, n_metrics):
        """
        Returns a formated string that helps to understand metrics bias.

        Parameters
        ----------
        metrics_biased: pd.Series
            A series than contains if biased or not for each metrics
        attr: str
            Protected attribute to inspect 
        target_value: str
            Current target values
        n_metrics: int
            Number of total metrics

        Returns
        -------
        str:
            formated string
        """
        n_biased = np.sum(metrics_biased)
        biased = 'not ' if n_biased == 0 else ''
        return (f"For this target value ({target_value}) regarding the '{attr}' attribute {n_biased} for the {n_metrics} bias metrics are/is biased so " +
                f"you can considered that the dataset is {biased}biased.")

    def insight_one_attr(self, attr):
        """
        Returns a formated dictionary that helps to understand metrics bias.
        
        The dictionary will have target value(s) as keys and formated string insight as
        values.

        Parameters
        ----------
        attr: str
            Protected attribute to inspect 

        Returns
        -------
        dict:
            Dictionary with target value(s) as keys and insight details as values
        """
        metrics = self.get_bias_metrics(attr=attr)
        metrics_biased = self.__metrics_is_biased(attr=attr)

        if self.favorable_label is not None:
            metrics = metrics.loc[self.favorable_label]

        values = metrics.index if type(
            metrics) is not pd.Series else [metrics.name]
        insight = {}

        if self.bias_type == 'dataset':
            for val in values:
                insight[val] = self.__insight_str(
                    metrics_biased.loc[val], attr=attr, target_value=val, n_metrics=2)

            return insight

        elif self.bias_type == 'model':
            for val in values:
                insight[val] = self.__insight_str(
                    metrics_biased.loc[val], attr=attr, target_value=val, n_metrics=5)

            return insight

        else:
            return None

    def insight(self, attr=None):
        """
        Returns a formated dictionary that helps to understand metrics bias.
        
        The dictionary will have attributes as keys and formated dictionnary insight as
        values.

        Parameters
        ----------
        attr: str (default None)
            Protected attribute to inspect (if None returns bias for all attributes)

        Returns
        -------
        dict:
            Dictionary with attributes as keys and insight details as values
        """
        if attr is None:
            attr_dict = {}
            for attr in self.protected_attributes:
                attr_dict[attr] = self.insight_one_attr(attr=attr)
        else:
            attr_dict = {attr: self.insight_one_attr(attr=attr)}
        return attr_dict

    def set_bias_type(self, bias_type):
        """
        Set the bias type of the Bias metric class.
        Can only be dataset or model.

        Parameters
        ----------
        bias_type: str
            Bias type of the object : only 'dataset' or 'model'
        """
        if bias_type not in ['dataset', 'model']:
            raise ValueError("bias_type should be 'dataset' or 'model'")

        self.bias_type = bias_type

    def plot_bias(self, attr=None, target_value=None):
        """
        Display a custom matplotlib graphic to show if a protected attribute is biased or not

        Current dataset bias metrics : 

        - Disparate impact 
        - Statistical parity difference

        Current model bias metrics :

        - Disparate impact 
        - Statistical parity difference
        - Equal opportunity difference
        - Average abs odds difference
        - Theil index

        Parameters
        ----------
        attr: str (default None)
            Protected attribute to inspect (if None display bias for all attributes)
        target_value: str (default None)
            Specific value of the target (if None display bias for all target values)
            It's usefull to avoid repetition in case of a binary classifier
        """
        if self.bias_type is None:
            raise ValueError(
                'Please set a bias type on this BiasMetric object (dataset or model)')

        if (target_value is None) and (self.favorable_label is not None):
            target_value = self.favorable_label
        if attr != None:
            plots.plot_bias_metrics(
                protected_attr=self.protected_attributes[attr],
                bias_type=self.bias_type,
                target_value=target_value)
        else:
            for attr in self.protected_attributes:
                display(Markdown(''))
                plots.plot_bias_metrics(
                    protected_attr=self.protected_attributes[attr],
                    bias_type=self.bias_type,
                    target_value=target_value)



import pandas as pd

import matplotlib.pyplot as plt
import seaborn as sns
from IPython.display import display, Markdown

from transparentai import plots

from sklearn.metrics import accuracy_score
from sklearn.metrics import f1_score
from sklearn.metrics import precision_score
from sklearn.metrics import recall_score


def plot_classification_scores(scores, y_true, y_proba):
    """
    Display different charts for all the metrics : 
    - a dataframe for accuracy, f1, precision, recall & roc_auc
    - confusion matrix
    - ROC curve 
    - Probality distribution

    Parameters
    ----------
    scores: pd.DataFrame
        scores dataframe attribute of a `ClassificationModel` object
    y_true: array-like
        true labels
    y_proba: array
        Target scores, can either be probability estimates of the positive class, 
        confidence values, or binary decisions.
    """
    n_classes = len(y_proba[0])
    fig = plt.figure(figsize=(15, 10))

    n_rows = 1 + n_classes
    ax = plt.subplot(int(f'221'))
    plot_confusion_matrix(matrix=scores['confusion_matrix'])

    ax = plt.subplot(int(f'222'))
    plot_roc_curve(roc_curve=scores['roc_curve'],
                   roc_auc=scores['roc_auc'], n_classes=n_classes)

    ax = plt.subplot(int(f'212'))
    plot_class_distribution(
        y_true=y_true, y_proba=y_proba, n_classes=n_classes)

    plots.plot_or_save(fname='classification_scores_plot.png')


def plot_confusion_matrix(matrix):
    """
    Show confusion matrix.

    Parameters
    ----------
    matrix: array
        confusion_matrix metrics result
    """
    sns.heatmap(matrix,
                cmap='Blues',
                square=True,
                fmt='d',
                annot=True)
    plt.xlabel('Predicted')
    plt.ylabel('Real')
    plt.title('Confusion Matrix')


def plot_roc_curve(roc_curve, roc_auc, n_classes):
    """
    Show a roc curve plot with roc_auc score on the legend.

    Parameters
    ----------
    roc_curve: array
        roc_curve metrics result for each class
    roc_auc: array
        roc_auc metrics result for each class       
    n_classes: int
        Number of classes
    """
    fpr, tpr = dict(), dict()

    for v in list(roc_curve.keys()):
        fpr[v] = roc_curve[v][0]
        tpr[v] = roc_curve[v][1]
    lw = 2

    colors = sns.color_palette("colorblind", n_classes)
    for v, color in zip(list(roc_auc.keys()), colors):
        n = 1 if n_classes == 1 else v
        plt.plot(fpr[v], tpr[v], color=color, lw=lw,
                 label='ROC curve of class {0} (area = {1:0.2f})'
                 ''.format(n, roc_auc[v]))
        if n_classes <= 2:
            break

    plt.plot([0, 1], [0, 1], color='navy', lw=lw, linestyle='--')
    plt.xlim([0.0, 1.0])
    plt.ylim([0.0, 1.05])
    plt.xlabel('False Positive Rate')
    plt.ylabel('True Positive Rate')
    plt.title('Receiver Operating Characteristic curve')
    plt.legend(loc="lower right")


def plot_class_distribution(y_true, y_proba, n_classes):
    """
    Show class distribution using seaborn `distplot`_ function.
    if n_classes is 2 then display probability for class 1 only.
    So class 0 probabilities should tend towards 0 and class 1 towards 1.

    else it displays all probabilities for current class so all 
    probabilities should tend towards 1.

    .. _distplot: https://seaborn.pydata.org/generated/seaborn.distplot.html

    Parameters
    ----------
    y_true: array-like
        true labels
    y_proba: array
        Target scores, can either be probability estimates of the positive class, 
        confidence values, or binary decisions.
    n_classes: int
        Number of classes
    """
    df = pd.DataFrame()
    df['y_true'] = y_true
    for idx, v in enumerate(list(df['y_true'].unique())):
        df[v] = [p[idx] for p in y_proba]

    if n_classes == 2:
        df[0] = df[1]

    colors = sns.color_palette("colorblind", n_classes)
    for v in df['y_true'].unique():
        kde = df[df['y_true'] == v][v].nunique() > 1
        sns.distplot(df[df['y_true'] == v][v], label=v, kde=kde)
        plt.legend(loc=0)

    plt.xlabel('Predicted probabilities')


def plot_curve(df, var):
    """
    Plot one curve based on a dataframe with a specific variable

    Parameters
    ----------
    df: pd.DataFrame
        Dataframe to inspect
    var: str
        Variable name inside dataframe
    """
    lw = 2
    plt.plot(df.index, df[var], color='navy', lw=lw, label=var)
    plt.xlabel('threshold')
    plt.ylabel(var)
    plt.legend(loc="lower right")


def compare_threshold_predictions(threshold_df, y_true):
    """
    Display curves for four metrics by threshold on the x axis :
    - accuracy
    - f1 score
    - precision
    - recall

    Parameters
    ----------
    threshold_df: pd.DataFrame
        dataframe with different threshold on columns
    y_true: array like
        true labels        
    """
    df = list()
    for threshold in threshold_df:
        accuracy = accuracy_score(y_true, threshold_df[threshold])
        f1 = f1_score(y_true, threshold_df[threshold])
        precision = precision_score(y_true, threshold_df[threshold])
        recall = recall_score(y_true, threshold_df[threshold])
        df.append([accuracy, f1, precision, recall])

    df = pd.DataFrame(df,
                      columns=['accuracy', 'f1_score',
                               'precision_score', 'recall_score'],
                      index=threshold_df.columns
                      )

    scores = df.columns
    fig = plt.figure(figsize=(15, 10))

    for i in range(0, 4):
        ax = fig.add_subplot(2, 2, i+1)
        plot_curve(df=df, var=scores[i])

    plots.plot_or_save(fname='compare_threshold_predictions_plot.png')



import pandas as pd
import numpy as np
from IPython.display import display, Markdown
from abc import abstractmethod

import transparentai.models.models_plots as plots
import transparentai.utils as utils


class BaseModel():
    """    
    Base class for different model types. 

    It allows to inspect a model based on a model which has a `predict` and 
    `predict_proba` (if classification) functions.

    It could help you to explore your model performance and validate or not your model.

    Attributes
    ----------
    X: pd.DataFrame
        X data that were used to get the predictions
    y_preds: np.array or pd.Series
        Predictions using X parameters
    y_true: np.array or pd.Series (optional)
        Real output
    model: 
        object with at least a `predict` function that returns
        a series or np.array
    scores_dict: dict
        Dictionary with metric name as keys and score value
        as values
    model_type: str
        String that indicates what model type it is.
        Can be 'classification' or 'regression'
    """

    scores_dict = None
    model_type = None

    def __init__(self, model=None, X=None, y=None, y_preds=None):
        """
        Parameters
        ----------
        model:
            a classifier model that have a `predict` and 
            `predict_proba` (if classification) functions
        X: pd.DataFrame (optional)
            X data that will be used to get the predictions
        y: np.array or pd.Series (optional)
            Real output
        y_preds: np.array or pd.Series (optional)
            Predictions using X parameters if model is not set
        """
        self.model = model
        self.X = X
        self.y_true = y
        self.y_preds = y_preds

    def _scores_to_json(self, scores_names):
        """
        Formats scores_dict attributes to be read as a json object.

        If 'roc_auc' is in the keys then it computes the score mean
        for the different target values 

        Returns
        -------
        dict:
            Scores dict formated for json
        """
        scores_json = {}
        for k, v in self.scores_dict.items():
            if k in scores_names:
                if k == 'roc_auc':
                    scores_json[k] = np.mean([val for label, val in v.items()])
                else:
                    scores_json[k] = v
        return scores_json

    def save_scores(self, fname):
        """
        Saves metric scores dictionary to a json file.

        Parameters
        ----------
        fname: str 
            string of the file path (including filename)
        """
        scores_json = self.scores_to_json()
        utils.save_dict_to_json(obj=scores_json, fname=fname)

    def display_scores(self):
        """
        Display current scores computed by `compute_scores()` function.

        Raises
        ------
        ValueError:
            compute_scores() function has to be compute first.
        """
        if self.scores_dict is None:
            raise ValueError('Use compute_scores() function first.')

        scores_to_display = None
        if self.model_type == 'classification':
            scores_to_display = ['accuracy', 'f1',
                                 'precision', 'recall', 'roc_auc']
        elif self.model_type == 'regression':
            scores_to_display = ['MAE', 'MSE', 'RMSE', 'R2']

        scores = {k: v for k, v in self.scores_dict.items()
                  if k in scores_to_display}
        scores = pd.Series(scores).to_frame().T
        scores.index = ['score']

        display(scores)

    def plot_overall_scores(self, fun, preds):
        """
        Display different charts for all the metrics.

        Raises
        ------
        ValueError:
            compute_scores() function has to be compute first.
        """
        if self.scores_dict is None:
            raise ValueError('Use compute_scores() function first.')

        display(Markdown('### Overall model performance'))
        self.display_scores()

        fun(self.scores_dict, self.y_true, preds)

    # Asbtract methods

    @abstractmethod
    def compute_scores(self):
        return None

    @abstractmethod
    def scores(self):
        return None

    @abstractmethod
    def scores_to_json(self):
        return None

    @abstractmethod
    def plot_scores(self):
        return None



__all__ = [
    'plot_or_save', 
    'plot_stack',
    'plot_stack_bar',
    'plot_boxplot_cat_num_var',
    'plot_correlation_matrix',
    'plot_gauge',
    'plot_bar',
    'save_plot',
    'save_dir',
    'verbose'
    ]

from transparentai.plots.plots import *


import os
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
import matplotlib.image as mpimg
import seaborn as sns
import numpy as np
from IPython.display import display, Markdown

from transparentai.fairness.protected_attribute import ProtectedAttribute
import transparentai.utils as utils
from transparentai import plots

bias_colors = ['#3498db', '#ecf0f1']
predicted_colors = ['#bdc3c7', '#2980b9']


def plot_text_center(ax, text, fontsize=18):
    """
    Display text at the center of an ax from a matplotlib figure.

    Parameters
    ----------
    ax: plt.axes.Axes
        ax where to set centered text
    text: str
        text to display
    fontsize: int (default 18)
        font size of the text
    """
    left, width = .25, .5
    bottom, height = .25, .5
    right = left + width
    top = bottom + height

    ax.text(0.5 * (left + right), 0.5 * (bottom + top), text,
            horizontalalignment='center',
            verticalalignment='center',
            fontsize=fontsize,
            transform=ax.transAxes)

    ax.axis('off')

# TODO : refactor text center & left


def plot_text_left(ax, text, fontsize=18):
    """
    Display text at the left of an ax from a matplotlib figure.

    Parameters
    ----------
    ax: plt.axes.Axes
        ax where to set centered text
    text: str
        text to display
    fontsize: int (default 18)
        font size of the text
    """
    left, width = 0, 0
    bottom, height = .25, .5
    right = left + width
    top = bottom + height

    ax.text(0.5 * (left + right), 0.5 * (bottom + top), text,
            horizontalalignment='center',
            verticalalignment='center',
            fontsize=fontsize, ha='left', multialignment='left',
            transform=ax.transAxes)

    ax.axis('off')


def plot_percentage_bar_man_img(ax, freq, spacing=0):
    """
    Display a percentage bar plot with custom images. only 10 images
    will be plotted, and each image has only one color so if the frequency is
    0.66 then it rounded up (0.7) and print 7 blue men and 3 grey.

    Parameters
    ----------
    ax: plt.axes.Axes
        ax where to plot the image barplot
    freq: str
        frequency to plot with man image
    spacing: int (default 0)
        spacing between images
    """
    heights = [70 for i in range(0, 10)]

    for i, height in enumerate(heights):
        img_path = os.path.abspath(os.path.join(
            os.path.dirname(__file__), '../src/img/'))
        img_path += '/_man_blue.png' if i < round(
            freq, 1)*10 else '/_man_grey.png'
        img = mpimg.imread(img_path)
        # AR = img.shape[1] / img.shape[0]
        AR = 135 / 256
        width = height * AR
        left = width*i + spacing*i
        right = left + width
        ax.imshow(img, extent=[left, right, 0, height])

    ax.set_xlim(0, right)
    ax.set_ylim(0, max(heights)*1.1)
    ax.axis('off')


def plot_pie(ax, freq, annot=False):
    """
    Display a pie chart from a frequency.

    This function uses bias_colors variable.

    Parameters
    ----------
    ax: plt.axes.Axes
        ax where to plot the pie chart
    freq: str
        frequency for the pie chart
    annot: bool (default False)
        add percentage annotation in the pie chart
    """
    sizes = [freq, 1-freq]
    autopct = '%1.0f%%' if annot else ''

    ax.pie(sizes, labels=['', ''], colors=bias_colors, autopct=autopct,
           shadow=True, startangle=130)
    ax.axis('equal')


def plot_protected_attr_row(axes, protected_attr, target_value, privileged=True):
    """
    Display a row of graphics for a protected attribute.
    privileged has to be True or False and then the function compute
    the frequency for this population.

    `axes` need 4 axes : 

    1. the percentage image plot (`plot_percentage_bar_man_img()`)
    2. an equal symbol
    3. pie chart of frequency
    4. text of the calculus

    Parameters
    ----------
    axes: list
        Axes for this row
    protected_attr:
        Protected attribute to inspect
    target_value: str
        Specific value of the target    
    privileged: bool (default True)
        Boolean prescribing whether to
        condition this metric on the `privileged_groups`, if `True`, or
        the `unprivileged_groups`, if `False`. Defaults to `None`
        meaning this metric is computed over the entire dataset.
    """
    n_total = protected_attr.num_instances(privileged=privileged)
    n = protected_attr.num_spec_value(
        target_value=target_value, privileged=privileged)
    freq = n / n_total
    title = 'Privileged' if privileged else 'Unprivileged'

    plot_percentage_bar_man_img(ax=axes[0], freq=freq)
    axes[0].set_title(title, loc='left', fontsize=18)
    plot_text_center(ax=axes[1], text='=', fontsize=25)
    plot_pie(ax=axes[2], freq=freq)

    text1 = '%.2f%%' % (freq*100)
    text2 = f'\n\n\n\n({n} / {n_total})'
    plot_text_center(ax=axes[3], text=text1, fontsize=32)
    plot_text_center(ax=axes[3], text=text2, fontsize=14)


def setup_bottom_line(ax):
    """
    Removes all axis except the bottom one.

    Code inspired from matplotlib.org_.

    .. _matplotlib.org: https://matplotlib.org/3.1.1/gallery/ticks_and_spines/tick-locators.html#sphx-glr-gallery-ticks-and-spines-tick-locators-py

    Parameters
    ----------
    ax: plt.axes.Axes
        ax to update
    """
    ax.spines['right'].set_color('none')
    ax.spines['left'].set_color('none')
    ax.spines['top'].set_color('none')
    ax.set_yticks([])
    ax.set_xticks([])


def set_legend_protected_attr(fig, axes, protected_attr, target_value):
    """
    Display title and legend of the bias graphic.

    Parameters
    ----------
    fig:
        Figure object
    axes: list
        Axes for this row
    protected_attr:
        Protected attribute to inspect
    target_value: str
        Specific value of the target  
    """
    target = protected_attr.target
    attr_name = protected_attr.name

    text = f'Focus on {attr_name} for {target} is {target_value}'
    unp_text = r'$\bf{Unprivileged}$' + \
        f' means {protected_attr.get_unprivileged_values()}'
    pri_text = r'$\bf{Privileged}$' + \
        f' means {protected_attr.get_privileged_values()}'

    setup_bottom_line(axes[0])

    axes[0].text(0.0, 0.9, text, fontsize=22,
                 transform=axes[0].transAxes, fontweight='bold')
    axes[0].text(0.0, 0.5, unp_text, fontsize=16,
                 transform=axes[0].transAxes, fontweight='normal')
    axes[0].text(0.0, 0.1, pri_text, fontsize=16,
                 transform=axes[0].transAxes, fontweight='normal')

    text_blue = f'{target} is {target_value}'
    text_grey = f'{target} is not {target_value}'
    fig.legend(labels=[text_blue, text_grey], fontsize=18)


def get_metric_text(protected_attr, target_value, metric_name, bias_type='dataset'):
    """
    Return a text value for a specific metric

    Parameters
    ----------
    protected_attr:
        Protected attribute to inspect
    target_value: str
        Specific value of the target  
    metric_name: str
        Metric's name 
    bias_type: str (default: 'dataset')
        Bias type (can only be 'dataset' or 'model')

    Returns
    -------
    str
        String of the metric calculus

    Raises
    ------
    ValueError:
        `bias_type` parameter has to be 'dataset' or 'model'
    """
    if bias_type not in ['dataset','model']:
        raise ValueError("bias_type has to be 'dataset' or 'model'")

    metrics = protected_attr.metrics
    value = metrics.loc[target_value, metric_name]

    if bias_type == 'dataset':
        freq_unpr = protected_attr.get_freq(target_value, privileged=False)
        freq_priv = protected_attr.get_freq(target_value, privileged=True)

        if metric_name == 'Disparate impact':
            formula = r'$\frac{Pr(Y = %s\ |\ D = unprivileged)}{Pr(Y = %s\ |\ D = privileged)}$' % (
                target_value, target_value)
            return formula + '\n'+r'$ = \frac{%.2f}{%.2f}$ = %.2f' % (freq_unpr, freq_priv, value)

        elif metric_name == 'Statistical parity difference':
            formula = r'$Pr(Y = %s\ |\ _{D = unprivileged}) - Pr(Y = %s\ |\ _{D = privileged})$' % (
                target_value, target_value)

            return formula + '\n' + r'= %.2f - %.2f = %.2f' % (freq_unpr, freq_priv, value)

    elif bias_type == 'model':
        freq_unpr = protected_attr.get_freq(
            target_value, privileged=False, predictions=True)
        freq_priv = protected_attr.get_freq(
            target_value, privileged=True, predictions=True)

        if metric_name == 'Disparate impact':
            formula = r'$\frac{Pr(\hat{Y} = %s\ |\ D = unprivileged)}{Pr(\hat{Y} = %s\ |\ D = privileged)}$' % (
                target_value, target_value)
            return formula + '\n'+r'$ = \frac{%.2f}{%.2f}$ = %.2f' % (freq_unpr, freq_priv, value)

        elif metric_name == 'Statistical parity difference':
            formula = r'$Pr(\hat{Y} = %s\ |\ _{D = unprivileged}) - Pr(\hat{Y} = %s\ |\ _{D = privileged})$' % (
                target_value, target_value)

            return formula + '\n' + r'= %.2f - %.2f = %.2f' % (freq_unpr, freq_priv, value)

        elif metric_name == 'Equal opportunity difference':
            TPR_unpriv = protected_attr.true_positive_rate(
                target_value, privileged=False)
            TPR_priv = protected_attr.true_positive_rate(
                target_value, privileged=True)

            formula = '$TPR_{unprivileged} - TPR_{privileged}$'
            return formula + '\n = %.2f - %.2f = %.2f' % (
                TPR_unpriv, TPR_priv, value)

        elif metric_name == 'Average abs odds difference':
            TPR_unpriv = protected_attr.true_positive_rate(
                target_value, privileged=False)
            TPR_priv = protected_attr.true_positive_rate(
                target_value, privileged=True)
            FPR_unpriv = protected_attr.false_positive_rate(
                target_value, privileged=False)
            FPR_priv = protected_attr.false_positive_rate(
                target_value, privileged=True)

            formula = r'$\frac{1}{2}\left[|FPR_{unprivileged} - FPR_{privileged}| + |TPR_{unprivileged} - TPR_{privileged}|\right]$'
            return formula + '\n' + r'= $\frac{1}{2}\left[|%.2f - %.2f| + |%.2f - %.2f|\right]$ = %.2f' % (
                FPR_unpriv, FPR_priv, TPR_unpriv, TPR_priv, value)

        elif metric_name == 'Theil index':
            formula = r'$\frac{1}{n}\sum_{i=1}^n\frac{b_{i}}{\mu}\ln\frac{b_{i}}{\mu}$'
            return formula + ' = %.2f' % (value)

    return 'no metric selected'


def plot_bias_metric(axes, protected_attr, target_value, metric, bias_type='dataset'):
    """
    Display a row of graphics for a specific bias metric.

    Parameters
    ----------
    axes: list
        Axes for this row
    protected_attr: transparentai.fairness.ProtectedAttribute
        Protected attribute to inspect
    target_value: str
        Specific value of the target    
    metric: str
        Metric's name
    bias_type: str (default: 'dataset')
        Bias type (can only be 'dataset' or 'model')

    Raises
    ------
    ValueError:
        `bias_type` parameter has to be 'dataset' or 'model'
    """
    if bias_type not in ['dataset','model']:
        raise ValueError("bias_type has to be 'dataset' or 'model'")
    value = protected_attr.metrics.loc[target_value, metric]
    goal = utils.get_metric_goal(metric=metric)

    text = get_metric_text(protected_attr, target_value,
                           metric_name=metric, bias_type=bias_type)
    plot_text_left(ax=axes[0], text=text, fontsize=19)
    axes[0].set_title(metric, loc='left', fontsize=22, fontweight='bold')
    axes[0].axis('off')

    color = 'r' if (value < goal-0.2) or (value > goal+0.2) else 'g'
    plots.plot_gauge(ax=axes[1], value=value, goal=goal, bar_color=color)
    axes[1].set_title(
        f'Considered not biased between {goal-0.2} and {goal+0.2}', loc='left', fontsize=17)


def plot_bar_freq_predicted(axes, protected_attr, target_value, privileged=True):
    """
    Display a row with a bar plot for a specific `ProtectedAttribute`.
    The graph represents frequency of predicted 0 versus predicted 1.

    Parameters
    ----------
    axes: list
        Axes for this row
    protected_attr: transparentai.fairness.ProtectedAttribute
        Protected attribute to inspect
    target_value: str
        Specific value of the target    
    privileged: bool (default True)
        Boolean prescribing whether to
        condition this metric on the `privileged_groups`, if `True`, or
        the `unprivileged_groups`, if `False`. Defaults to `None`
        meaning this metric is computed over the entire dataset.
    """
    freq_1 = (protected_attr.get_freq(target_value,
                                      privileged, predictions=True)*100).round(2)
    freq_0 = 100 - freq_1
    perf = protected_attr.performances(target_value, privileged)
    TP, FP, TN, FN = perf['TP'], perf['FP'], perf['TN'], perf['FN']
    TPR = TP / (TP+FN)
    FPR = FP / (TN+FP)
    priv_text = 'Privileged' if privileged else 'Unprivileged'

    rects = axes[0].barh(['', ''], [freq_0, freq_1], color=predicted_colors,
                         left=[0, freq_0], label=['', ''])

    for r in rects:
        h, w, x, y = r.get_height(), r.get_width(), r.get_x(), r.get_y()
        if h == 0:
            continue

        axes[0].annotate(str(w)+'%', xy=(x+w/2, y+h/2), xytext=(0, 0),
                         textcoords="offset points", fontsize=16,
                         ha='center', va='center',
                         color='white', weight='bold', clip_on=True)

    axes[0].text(0, 0.7, r"$FPR_{%s} = \frac{FP}{TN+FP} = \frac{%i}{%i+%i} = %.2f$" % (
        priv_text, FP, TN, FP, FPR), fontsize=20)
    axes[0].text(100, 0.7, r"$TPR_{%s} = \frac{TP}{TP+FN} = \frac{%i}{%i+%i} = %.2f$" % (
        priv_text, TP, TP, FN, TPR), fontsize=20, ha='right')

    axes[0].axis('off')
    axes[0].set_xlim(0, 100)
    axes[0].text(
        0, -0.6, f"{priv_text} predicted not {target_value}", fontsize=16)
    axes[0].text(
        100, -0.6, f"{priv_text} predicted {target_value}", fontsize=16, ha='right')


def plot_protected_attr_row_model(axes, protected_attr, target_value, privileged=True):
    """
    Display a row of graphics for a protected attribute.
    privileged has to be True or False and then the function compute
    the frequency for this population.

    axes need 4 axes : 

    1. the percentage image plot (`plot_percentage_bar_man_img()`) for rows predicted as not target value  (aka negatives)
    2. pie chart of frequency for negatives
    3. pie chart of frequency for positives
    4. the percentage image plot (`plot_percentage_bar_man_img()`) for rows predicted as target value (aka positives)

    Parameters
    ----------
    axes: list
        Axes for this row
    protected_attr:
        Protected attribute to inspect
    target_value: str
        Specific value of the target    
    privileged: bool (default True)
        Boolean prescribing whether to
        condition this metric on the `privileged_groups`, if `True`, or
        the `unprivileged_groups`, if `False`. Defaults to `None`
        meaning this metric is computed over the entire dataset.
    """
    perf = protected_attr.performances(target_value, privileged)
    TP, FP, TN, FN = perf['TP'], perf['FP'], perf['TN'], perf['FN']
    freq_neg = FN / (FN+TN)
    freq_pos = TP / (FP+TP)

    priv_text = 'Privileged' if privileged else 'Unprivileged'

    plot_percentage_bar_man_img(ax=axes[0], freq=freq_neg)
    axes[0].set_title(f'{priv_text} predicted not {target_value}',
                      loc='left', fontsize=18, fontweight='bold')
    axes[0].text(0, -18, f"FN: {FN}", fontsize=16)
    axes[0].text(300, -18, f"TN: {TN}", fontsize=16)

    plot_pie(ax=axes[1], freq=freq_neg, annot=True)
    plot_pie(ax=axes[2], freq=freq_pos, annot=True)
    plot_percentage_bar_man_img(ax=axes[3], freq=freq_pos)
    axes[3].set_title(f'{priv_text} predicted {target_value}',
                      loc='right', fontsize=18, fontweight='bold')
    axes[3].text(0, -18, f"TP: {TP}", fontsize=16)
    axes[3].text(300, -18, f"FP: {FP}", fontsize=16)


def set_metric_title(axes, text, fontsize=22):
    """
    Display bias metrics title.

    Parameters
    ----------
    axes: list
        Axes for this row 
    text: str
        Text for the title
    fontsize: int (default 22)
        font size of the text
    """
    setup_bottom_line(axes[0])
    axes[0].text(0.0, 0.2, text, fontsize=fontsize,
                 transform=axes[0].transAxes, fontweight='bold')


def generate_gridspec_row(fig, gs, rstart, rsize=1, n_blocks=1, b_sizes=[10]):
    """
    Return a list of axes for a gridpsec graphic.
    You can use this function with the following logic :

    if you want a row of 3 blocks of size 2,2,6 at the 3rd row :

    >>> generate_gridspec_row(fig, gs, rstart=3, rsize=1, n_blocks=3, b_sizes=[2,2,6])

    Parameters
    ----------
    fig:
        Figure
    gs:
        returns of fig.add_gridspec() function
    rstart: int
        Row start
    rsize: int (default 1)
        Size of the row (number of gridspec rows)
    n_blocks: int (default 1)
        Number of different axes for this row
    b_sizes: list (default [10])
        Sizes of the blocks

    Returns
    -------
    list:
        list of axes for this row
    """
    rows = list()
    b_end = 0
    for i in range(0, n_blocks):
        b_start = b_end
        b_end += b_sizes[i]
        rows.append(fig.add_subplot(gs[rstart:rstart+rsize, b_start:b_end]))
    return rows


def plot_dataset_bias_metrics(protected_attr, target_value):
    """
    Display matplotlib graphics with different information about
    bias inside a dataset.

    Parameters
    ----------
    protected_attr:
        Protected attribute to inspect
    target_value: str
        Specific value of the target        
    """
    fig = plt.figure(constrained_layout=True, figsize=(16, 14))
    gs = fig.add_gridspec(8, 10)

    # Unpriviliged plot
    axes = generate_gridspec_row(
        fig, gs, rstart=1, rsize=2, n_blocks=4, b_sizes=[5, 1, 2, 2])
    plot_protected_attr_row(axes, protected_attr,
                            target_value, privileged=False)

    # Priviliged plot
    axes = generate_gridspec_row(
        fig, gs, rstart=3, rsize=2, n_blocks=4, b_sizes=[5, 1, 2, 2])
    plot_protected_attr_row(axes, protected_attr,
                            target_value, privileged=True)

    # Plot title and legend
    axes = generate_gridspec_row(
        fig, gs, rstart=0, rsize=1, n_blocks=1, b_sizes=[10])
    set_legend_protected_attr(fig, axes, protected_attr, target_value)

    # Metric title
    axes = generate_gridspec_row(
        fig, gs, rstart=5, rsize=1, n_blocks=1, b_sizes=[10])
    set_metric_title(axes, text='Dataset Bias metrics')

    # Metrics plots
    metrics = ['Disparate impact', 'Statistical parity difference']
    for idx, metric in enumerate(metrics):
        axes = generate_gridspec_row(
            fig, gs, rstart=6+idx, rsize=1, n_blocks=2, b_sizes=[4, 6])
        plot_bias_metric(axes, protected_attr, target_value,
                         metric=metric, bias_type='dataset')

    plots.plot_or_save(fname='dataset_bias_metrics_plot.png')


def plot_model_bias_metrics(protected_attr, target_value):
    """
    Display a matplotlib graphics with differents informations about
    bias inside a dataset.

    Parameters
    ----------
    protected_attr:
        Protected attribute to inspect
    target_value: str
        Specific value of the target        
    """
    fig = plt.figure(constrained_layout=True, figsize=(16, 20))
    gs = fig.add_gridspec(13, 10)

    # Unpriviliged plot
    axes = generate_gridspec_row(
        fig, gs, rstart=1, rsize=2, n_blocks=4, b_sizes=[3, 2, 2, 3])
    plot_protected_attr_row_model(axes, protected_attr,
                                  target_value, privileged=False)
    axes = generate_gridspec_row(
        fig, gs, rstart=3, rsize=1, n_blocks=1, b_sizes=[10])
    plot_bar_freq_predicted(axes, protected_attr,
                            target_value, privileged=False)

    # Priviliged plot
    axes = generate_gridspec_row(
        fig, gs, rstart=4, rsize=2, n_blocks=4, b_sizes=[3, 2, 2, 3])
    plot_protected_attr_row_model(axes, protected_attr,
                                  target_value, privileged=True)
    axes = generate_gridspec_row(
        fig, gs, rstart=6, rsize=1, n_blocks=1, b_sizes=[10])
    plot_bar_freq_predicted(axes, protected_attr,
                            target_value, privileged=True)

    # Plot title and legend
    axes = generate_gridspec_row(
        fig, gs, rstart=0, rsize=1, n_blocks=1, b_sizes=[10])
    set_legend_protected_attr(fig, axes, protected_attr, target_value)

    # Metric title
    axes = generate_gridspec_row(
        fig, gs, rstart=7, rsize=1, n_blocks=1, b_sizes=[10])
    set_metric_title(axes, text='Model Bias metrics')

    # Metrics plots
    metrics = ['Disparate impact', 'Statistical parity difference',
               'Equal opportunity difference', 'Average abs odds difference',
               'Theil index']
    start = 8
    for idx, metric in enumerate(metrics):
        axes = generate_gridspec_row(
            fig, gs, rstart=idx+start, rsize=1, n_blocks=2, b_sizes=[4, 6])
        plot_bias_metric(axes, protected_attr, target_value,
                         metric=metric, bias_type='model')

    plots.plot_or_save(fname='model_bias_metrics_plot.png')


def plot_bias_metrics(protected_attr, bias_type, target_value=None):
    """
    Display a matplotlib graphics with different information about
    bias inside a dataset.

    Parameters
    ----------
    protected_attr:
        Protected attribute to inspect
    bias_type: str
        type of bias to plot ('dataset' or 'model')
    target_value: str (default None)
        Specific value of the target        
    """
    if bias_type not in ['dataset', 'model']:
        raise ValueError("bias_type should be 'dataset' or 'model'")

    func = plot_dataset_bias_metrics if bias_type == 'dataset' else plot_model_bias_metrics
    attr = protected_attr.name
    target = protected_attr.target

    if target_value is not None:
        func(protected_attr=protected_attr, target_value=target_value)
    else:
        for target_value in protected_attr.labels.unique():
            func(protected_attr=protected_attr, target_value=target_value)



import pandas as pd
from IPython.display import display, Markdown

from sklearn.metrics import accuracy_score
from sklearn.metrics import confusion_matrix
from sklearn.metrics import f1_score
from sklearn.metrics import precision_score
from sklearn.metrics import recall_score
from sklearn.metrics import roc_auc_score
from sklearn.metrics import roc_curve
from sklearn.preprocessing import label_binarize

import transparentai.models.models_plots as plots
from transparentai.models.base_model import BaseModel


class ClassificationModel(BaseModel):
    """    
    Class to inspect a classification model based on a model which has a `predict` and 
    `predict_proba` functions.
    It could help you to explore your model performance and validate or not your model.

    Example
    -------
    
    >>> from transparentai.models import ClassificationModel
    >>> model = ClassificationModel(model=clf)
    >>> model.compute_scores(X=X_test, y=y_test, threshold=0.5)

    If you want to analyse false positives or false negatives for example
    you can use the following code (also available for true positives and negatives)
    
    >>> FP_df = model.get_false_positives()
    >>> FN_df = model.get_false_negatives()

    For more details please see the `ClassificationModel for binary classification`_ or
    `ClassificationModel for multi labels classification`_ notebooks.

    .. _ClassificationModel for binary classification : https://github.com/Nathanlauga/transparentai/notebooks/example_ClassificationModel_binary_classification.ipynb
    .. _ClassificationModel for multi labels classification : https://github.com/Nathanlauga/transparentai/notebooks/example_ClassificationModel_multi_label_classification.ipynb

    Attributes
    ----------
    X: pd.DataFrame
        X data that were used to get the predictions
    y_preds: np.array or pd.Series
        Predictions using X parameters
    y_true: np.array or pd.Series (optional)
        Real output
    model: 
        object with at least a `predict` function that returns
        a series or np.array
    scores_dict: dict
        Dictionary with metric name as keys and score value
        as values
    model_type: str
        String that indicates what model type it is : 'classification'
    n_classes: int
        Number of classes inside y
    threshold_df: pd.DataFrame
        Dataframe with different probability thresholds as columns
        and their prediction (0 or 1), it can be use only for
        binary classification
    """

    def __init__(self, model=None, X=None, y=None, y_preds=None):
        """
        Parameters
        ----------
        model:
            a classifier model that have a `predict` and 
            `predict_proba` (if classification) functions
        X: pd.DataFrame (optional)
            X data that will be used to get the predictions
        y: np.array or pd.Series (optional)
            Real output
        y_preds: np.array or pd.Series (optional)
            Predictions using X parameters if model is not set
        """
        super().__init__(model=model, X=X, y=y, y_preds=y_preds)
        self.threshold_df = None
        self.y_proba = None
        self.model_type = 'classification'

        if (X is not None) & (y is not None) & (y_preds is not None):
            self.n_classes = len(list(set(y)))
            self.scores()

    def compute_scores(self, X=None, y=None, threshold=0.5):
        """
        Compute all predictions, probalities and scores
        if it's binary classifier you can custom the probability threshold

        Parameters
        ----------
        X: array-like of shape (n_samples, n_features)
            feature samples
        y: array-like of shape (n_samples,) or (n_samples, n_outputs)
            true labels for X.
        threshold: float (default 0.5)
            only for binary classifier, custome threshold probability 
            for the prediction

        Raises
        ------
        AttributeError:
            X has to be set in init part or in the parameters
        AttributeError:
            y_true has to be set in init part or in the parameters as y
        AttributeError:
            model attribute was not set at the init step
        """
        if (X is None) & (self.X is None):
            raise AttributeError('X is mandatory to compute scores')
        if (y is None) & (self.y_true is None):
            raise AttributeError('y is mandatory to compute scores')
        if self.model is None:
            raise AttributeError('model attribute was not set at the init step')

        self.X = X if X is not None else self.X
        self.y_true = y if y is not None else self.y_true
        self.n_classes = len(list(set(y)))

        self.y_proba = self.model.predict_proba(X)

        if self.n_classes == 2:
            self.y_preds = (self.y_proba[:, 1] >= threshold).astype(int)
            self.compute_threshold(X=X, y=y)
        else:
            self.y_preds = self.model.predict(X)

        self.scores()

    def scores(self):
        """
        Compute classification metrics scores based on `skearn Model evaluation`_
        Current metrics : accuracy, confusion_matrix, f1_score, precision, recall and roc_auc

        .. _skearn Model evaluation: https://scikit-learn.org/0.15/modules/model_evaluation.html
        """
        scores = {}
        scores['accuracy'] = accuracy_score(self.y_true, self.y_preds)
        scores['confusion_matrix'] = confusion_matrix(
            self.y_true, self.y_preds)

        average = 'micro' if self.n_classes != 2 else 'weighted'
        scores['f1'] = f1_score(self.y_true, self.y_preds, average=average)
        scores['precision'] = precision_score(
            self.y_true, self.y_preds, average=average)
        scores['recall'] = recall_score(
            self.y_true, self.y_preds, average=average)

        if self.y_proba is not None:
            roc_auc, roc_curves = dict(), dict()
            if self.n_classes > 2:
                y = label_binarize(self.y_true, classes=self.y_true.unique())
                for i, v in enumerate(list(self.y_true.unique())):
                    try:
                        roc_auc[int(v)] = roc_auc_score(y[:, i], self.y_proba[:, i])
                        roc_curves[int(v)] = roc_curve(y[:, i], self.y_proba[:, i])
                    except ValueError:
                        roc_auc[int(v)] = 1
                        roc_curves[int(v)] = [[1,1], [0,1]]
                        pass
            else:
                roc_auc[0] = roc_auc_score(self.y_true, self.y_proba[:, 1])
                roc_curves[0] = roc_curve(self.y_true, self.y_proba[:, 1])

            scores['roc_auc'] = roc_auc # [v for k, v in roc_auc.items()]
            scores['roc_curve'] = roc_curves # [v for k, v in roc_curves.items()]

        self.scores_dict = scores

    def get_true_positives(self):
        """
        Return true positives rows in the feature sample X

        Returns
        -------
        pd.DataFrame
            True positives rows 
        """
        return self.X[(self.y_preds == 1) & (self.y_true == 1)]

    def get_true_negatives(self):
        """
        Return true negatives rows in the feature sample X

        Returns
        -------
        pd.DataFrame
            True negatives rows 
        """
        return self.X[(self.y_preds == 0) & (self.y_true == 0)]

    def get_false_positives(self):
        """
        Return false positives rows in the feature sample X

        Returns
        -------
        pd.DataFrame
            False positives rows 
        """
        return self.X[(self.y_preds == 1) & (self.y_true == 0)]

    def get_false_negatives(self):
        """
        Return false negatives rows in the feature sample X

        Returns
        -------
        pd.DataFrame
            False negatives rows 
        """
        return self.X[(self.y_preds == 0) & (self.y_true == 1)]

    def plot_scores(self):
        """
        Display different charts for all the metrics : 

        - a dataframe for accuracy, f1, precision, recall & roc_auc
        - confusion matrix
        - ROC curve 
        - Probalities distribution
        """
        fun = plots.plot_classification_scores
        self.plot_overall_scores(fun=fun, preds=self.y_proba)

    def scores_to_json(self):
        """
        Formats scores_dict attributes to be read as a json object.

        If 'roc_auc' is in the keys then it computes the score mean
        for the different target values 

        Returns
        -------
        dict:
            Scores dict formated for json
        """
        scores_names = ['accuracy', 'f1', 'precision', 'recall', 'roc_auc']
        return self._scores_to_json(scores_names)

    def compute_threshold(self, X, y, start=0, end=1, step=0.05):
        """
        Compute all predictions, probalities and scores
        if it's binary classifier you can custom the probability threshold

        Parameters
        ----------
        X: array-like of shape (n_samples, n_features)
            feature samples
        y: array-like of shape (n_samples,) or (n_samples, n_outputs)
            true labels for X.
        start: int, (default 0)
            minimum probability to compare 
            (should be between 0 and 1 and less than end value)
        end: int, (default 1)
            minimum probability to compare 
            (should be between 0 and 1 and greater than start value)
        step: float, (default 0.05)
            value between each threshold steps

        Raises
        ------
        ValueError:
            This function can be only called for binary classification.
        ValueError:
            start has to be smaller than end
        ValueError:
            start has to be between 0 and 1
        ValueError:
            end has to be between 0 and 1
        """
        if self.n_classes > 2:
            raise ValueError(
                'This function can be only called for binary classification.')
        if start > end:
            raise ValueError('start has to be smaller than end')
        if (start < 0) or (start > 1):
            raise ValueError('start has to be between 0 and 1')
        if (end < 0) or (end > 1):
            raise ValueError('end has to be between 0 and 1')

        threshold = start
        threshold_df = pd.DataFrame()

        proba = self.model.predict_proba(X)

        while threshold <= end:
            threshold = round(threshold, 2)
            preds = (proba[:, 1] >= threshold).astype(int)

            threshold_df[threshold] = preds
            threshold += step

        self.threshold_df = threshold_df

    def plot_threshold(self):
        """
        Display curves for four metrics by threshold on x axis :

        - accuracy
        - f1 score
        - precision
        - recall

        Raises
        ------
        ValueError:
            This function can be only called for binary classification.
        ValueError:
            compute_threshold() function has to be compute first.
        """
        if self.n_classes > 2:
            raise ValueError(
                'This function can be only called for binary classification.')
        if self.threshold_df is None:
            raise ValueError('Use compute_threshold() function first.')

        display(Markdown(
            '## Proba threshold comparison for accuracy, f1 score, precision & recall'))

        plots.compare_threshold_predictions(self.threshold_df, self.y_true)



__all__ = [
    'load_adult',
    'load_iris',
    'load_boston',
    'StructuredDataset',
    'plot_missing_values',
    'plot_numerical_var',
    'plot_categorical_var',
    'plot_datetime_var',
    'display_meta_var',
    'plot_numerical_jointplot',
    'plot_one_cat_and_num_variables',
    'plot_correlation_matrix'
]

from transparentai.datasets.structured_dataset import StructuredDataset
from transparentai.datasets.datasets import *
from transparentai.datasets.datasets_plots import *



import pandas as pd
from transparentai.fairness.protected_attribute import ProtectedAttribute


class DatasetProtectedAttribute(ProtectedAttribute):
    """
    This class retrieves all informations on a protected attribute on a specific dataset.
    It computes automatically the `Disparate impact` and the `Statistical parity difference` to
    get some insight about bias in the dataset.

    This class is inspired by the BinaryLabelDatasetMetric_ class from aif360 module but it 
    depends on a unique attribute.

    .. _BinaryLabelDatasetMetric: https://aif360.readthedocs.io/en/latest/modules/metrics.html#binary-label-dataset-metric

    Attributes
    ----------
    name: str
        Attribute name (the column's name in original dataframe)
    target:
        Name of the label (or target) column
    favorable_label:
        A specific value that is has an advantage over the other(s) values
    privileged_values: list
        List with privileged values inside the column (e.g. ['Male'] for 'gender' attribute)  
    unprivileged_values: list
        List with unprivileged values inside the column (e.g. ['Female'] for 'gender' attribute)  
    values: pd.Series
        Serie with the same length than labels attribute with 0 or 1 values 
        1 if it correspond to a privileged value, 0 if not.
    labels: np.array 
        Array with original label values        
    crosstab: pd.DataFrame
        Crosstab of values and labels
    """

    def __init__(self, dataset, attr, privileged_values, favorable_label=None):
        """
        Parameters
        ----------
        dataset: pd.DataFrame
            Dataframe to inspect
        attr: str
            Name of the attribute to analyse (it has to be into dataset columns)
        privileged_values: list
            List with privileged values inside the column (e.g. ['Male'] for 'gender' attribute)  
        favorable_label:
            A specific value that is has an advantage over the other(s) values
        """
        super().__init__(dataset, attr, privileged_values, favorable_label)
        self.compute_dataset_bias_metrics()

    def num_spec_value(self, target_value, privileged=None):
        r"""
        Computes the number of a particular value,
        :math:`P = \sum_{i=1}^n \mathbb{1}[y_i = v]`,
        optionally conditioned on protected attributes.

        Parameters
        ----------
        target_value:
            Specific value of the target
        privileged (bool, optional): 
            Boolean prescribing whether to
            condition this metric on the `privileged_groups`, if `True`, or
            the `unprivileged_groups`, if `False`. Defaults to `None`
            meaning this metric is computed over the entire dataset.

        Returns
        -------
        int:
            Number of a specified target value of all data if privileged is None, privileged
            values if privileged is True and unprivileged values if False.
        """
        if privileged == None:
            return self.crosstab.loc['All', target_value]
        elif privileged:
            return self.crosstab.loc[1, target_value]
        else:
            return self.crosstab.loc[0, target_value]

    
    def get_freq(self, target_value, privileged=None):
        """
        Returns the frequency of a specified target value. 

        Parameters
        ----------
        target_value:
            Specific value of the target
        privileged (bool, default None): 
                Boolean prescribing whether to
                condition this metric on the `privileged_groups`, if `True`, or
                the `unprivileged_groups`, if `False`. Defaults to `None`
                meaning this metric is computed over the entire dataset.
                
        Returns
        -------
        float
            Frequency of a specified target value
        """
        n_total = self.num_instances(privileged=privileged)
        n = self.num_spec_value(
            target_value=target_value, privileged=privileged)

        return n / n_total

    def base_rate(self, target_value, privileged=None):
        """
        Computes the base rate, :math:`Pr(Y = 1) = P/(P+N)`, optionally
        conditioned on protected attributes.

        Parameters
        ----------
        target_value:
            Specific value of the target
        privileged (bool, optional): 
                Boolean prescribing whether to
                condition this metric on the `privileged_groups`, if `True`, or
                the `unprivileged_groups`, if `False`. Defaults to `None`
                meaning this metric is computed over the entire dataset.

        Returns
        -------
        float: 
            Base rate
        """
        return (self.num_spec_value(target_value=target_value, privileged=privileged)
                / self.num_instances(privileged=privileged))

    def disparate_impact(self, target_value):
        r"""
        Computes the disparate impact for a specific label value

        .. math::
           \frac{Pr(Y = v | D = \text{unprivileged})}
           {Pr(Y = v | D = \text{privileged})}

        Parameters
        ----------
        target_value:
            Specific label value for which it will compute
            the metric

        Returns
        -------
        float:
            Disparate impact bias metric
        """
        return self.ratio(self.base_rate, target_value=target_value)

    def statistical_parity_difference(self, target_value):
        r"""
        Computes the statistical parity difference for a specific label value

        .. math::
           Pr(Y = v | D = \text{unprivileged})
           - Pr(Y = v | D = \text{privileged})

        Parameters
        ----------
        target_value:
            Specific label value for which it will compute
            the metric

        Returns
        -------
        number:
            Statistical parity difference bias metric
        """
        return self.difference(self.base_rate, target_value=target_value)

    def compute_dataset_bias_metrics(self):
        """
        Computes automaticaly all dataset bias metrics for this
        protected attribute.
        """
        metrics_dict = {
            'Disparate impact': self.disparate_impact,
            'Statistical parity difference': self.statistical_parity_difference
        }

        self.compute_bias_metrics(metrics_dict=metrics_dict)



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()
fh.close()

setup(
    name='transparentai',
    version='0.1.1',
    description="Python tool to create or inspect a transparent and ethical AI.",
    license='MIT',
    author='Nathan LAUGA',
    author_email='nathan.lauga@protonmail.com',
    url='https://github.com/Nathanlauga/transparentai',
    packages=[pkg for pkg in find_packages(
    ) if pkg.startswith('transparentai')],
    include_package_data=True,
    install_requires=[
        'pandas',
        'numpy',
        'matplotlib',
        'seaborn',
        'scipy',
        'scikit-learn',
        'ipython',
        'shap'
    ],
    long_description_content_type="text/markdown",
    long_description=long_description,
    python_requires='>=3.5',
    classifiers=[
        "Topic :: Software Development :: Libraries :: Python Modules",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
    ],
    entry_points={

    },
)



import pandas as pd
import os


def load_adult():
    """
    Load Adult dataset.
    Source : https://archive.ics.uci.edu/ml/datasets/Adult
    """

    names = [
        'age',
        'workclass',
        'fnlwgt',
        'education',
        'education-num',
        'marital-status',
        'occupation',
        'relationship',
        'race',
        'gender',
        'capital-gain',
        'capital-loss',
        'hours-per-week',
        'native-country',
        'income',
    ]
    dtypes = {
        'workclass': 'category',
        'education': 'category',
        'marital-status': 'category',
        'occupation': 'category',
        'relationship': 'category',
        'race': 'category',
        'gender': 'category',
        'native-country': 'category',
        'income': 'category'
    }
    adult = pd.read_csv(
        os.path.join(os.path.dirname(__file__), 'data', 'adult.csv'),
        names=names,
        header=None,
        dtype=dtypes
    )
    return adult


def load_iris():
    """
    Load Iris dataset.
    Source : http://archive.ics.uci.edu/ml/datasets/Iris/
    """

    names = [
        'sepal length (cm)',
        'sepal width (cm)',
        'petal length (cm)',
        'petal width (cm)',
        'iris plant'
    ]
    dtypes = {
        'iris plant': 'category'
    }
    iris = pd.read_csv(
        os.path.join(os.path.dirname(__file__), 'data', 'iris.csv'),
        names=names,
        header=None,
        dtype=dtypes
    )
    return iris

def load_boston():
    """
    Load boston dataset 
    Source : https://archive.ics.uci.edu/ml/machine-learning-databases/housing/ 
    """
    boston = pd.read_csv(
        os.path.join(os.path.dirname(__file__), 'data', 'boston.csv')
    )
    return boston


__all__ = [
    'ModelExplainer',
    'plot_global_feature_influence',
    'plot_local_feature_influence'
]

from transparentai.explainer.model_explainer import ModelExplainer
from transparentai.explainer.explainer_plots import *



import pandas as pd
import numpy as np

from transparentai.fairness.protected_attribute import ProtectedAttribute


class ModelProtectedAttribute(ProtectedAttribute):
    """
    This class retrieves all informations on a protected attribute on a specific model.
    It computes automatically the `Disparate impact`, `Statistical parity difference`, 
    `Equal opportunity difference`, `Average abs odds difference` and `Theil index` to
    get some insight about bias in the model.

    This class is inspired by the ClassificationMetric_ class from aif360 module but it 
    depends on a unique attribute.

    .. _ClassificationMetric: https://aif360.readthedocs.io/en/latest/modules/metrics.html#classification-metric

    Attributes
    ----------
    name: str
        Attribute name (the column's name in original dataframe)
    target:
        Name of the label (or target) column
    favorable_label:
        A specific value that is has an advantage over the other(s) values
    preds:
        Array with predictions on the dataset  
    privileged_values: list
        List with privileged values inside the column (e.g. ['Male'] for 'gender' attribute)  
    unprivileged_values: list
        List with unprivileged values inside the column (e.g. ['Female'] for 'gender' attribute)  
    values: pd.Series
        Serie with the same length than labels attribute with 0 or 1 values 
        1 if it correspond to a privileged value, 0 if not.
    labels: np.array 
        Array with original label values        
    crosstab: pd.DataFrame
        Crosstab of values and labels
    preds_crosstab:
        Crosstab of values and predicted labels
    """

    def __init__(self, dataset, attr, privileged_values, preds, favorable_label):
        """
        Parameters
        ----------
        dataset: pd.DataFrame
            Dataframe to inspect
        attr: str
            Name of the attribute to analyse (it has to be into dataset columns)
        privileged_values: list
            List with privileged values inside the column (e.g. ['Male'] for 'gender' attribute)  
        preds:
            Array with predictions on the dataset  
        favorable_label:
            A specific value that is has an advantage over the other(s) values
        """
        super().__init__(dataset, attr, privileged_values, favorable_label)
        self.preds = preds
        self.preds_crosstab = pd.crosstab(
            self.values, self.preds, margins=True)
        self.compute_model_bias_metrics()

    def confusion_matrix(self, privileged=None):
        """
        Computes the confusion matrix on all data if privileged is None else
        on whether privileged sample or unprivileged one.

        Parameters
        ----------
        privileged (bool, optional):
            Boolean prescribing whether to
            condition this metric on the `privileged_groups`, if `True`, or
            the `unprivileged_groups`, if `False`. Defaults to `None`
            meaning this metric is computed over the entire dataset.

        Returns
        -------
        pd.DataFrame:
            Confusion matrix
        """
        perf_df = self.to_frame()
        perf_df['preds'] = self.preds
        if privileged is True:
            perf_df = perf_df[perf_df[self.name] == 1]
        elif privileged is False:
            perf_df = perf_df[perf_df[self.name] == 0]

        matrix = pd.crosstab(perf_df[self.target], perf_df['preds'])

        true = perf_df[self.target].values
        pred = perf_df['preds'].values

        for val in list(set(self.preds)):
            if val not in matrix.columns:
                matrix[val] = 0

        for val in matrix.columns.values:
            if val not in matrix.index.values:
                matrix.loc[val, :] = 0
                
        for val in matrix.index.values:
            if val not in matrix.columns.values:
                matrix.loc[:, val] = 0

        return matrix

    def performances(self, target_value, privileged=None):
        """
        Computes the performances on all data if privileged is None else
        on whether privileged sample or unprivileged one.

        Performances corresponds to TP, FP, TN and FN.

        Parameters
        ----------
        target_value:
            Specific label value for which it will compute the metric
        privileged (bool, optional):
            Boolean prescribing whether to
            condition this metric on the `privileged_groups`, if `True`, or
            the `unprivileged_groups`, if `False`. Defaults to `None`
            meaning this metric is computed over the entire dataset.

        Returns
        -------
        dict:
            dictionary with True Positives, False Positives,
            True Negatives and False Negatives
        """
        other_values = [v for v in list(set(self.preds)) if v != target_value]
        matrix = self.confusion_matrix(privileged=privileged)

        TP = matrix.loc[target_value, target_value]
        FP = sum(matrix.loc[other_values, target_value].values)
        TN = sum(sum(matrix.loc[other_values, other_values].values))
        if type(target_value) == type(int()):
            FN = sum(matrix.iloc[target_value, other_values].values)
        else:
            FN = sum(matrix.loc[target_value, other_values].values)
        return dict(
            TP=TP, FP=FP, TN=TN, FN=FN
        )

    def num_spec_value(self, target_value, privileged=None, predictions=False):
        r"""
        Compute the number of a particular value,
        :math:`P = \sum_{i=1}^n \mathbb{1}[y_i = v]`,
        optionally conditioned on protected attributes.

        Parameters
        ----------
        target_value:
            Specific value of the target
        privileged (bool, optional): 
            Boolean prescribing whether to
            condition this metric on the `privileged_groups`, if `True`, or
            the `unprivileged_groups`, if `False`. Defaults to `None`
            meaning this metric is computed over the entire dataset.

        Returns
        -------
        int:
            Number of a specified target value of all data if privileged is None, privileged
            values if privileged is True and unprivileged values if False.
        """
        crosstab = self.crosstab if not predictions else self.preds_crosstab
        if privileged == None:
            return crosstab.loc['All', target_value]
        elif privileged:
            return crosstab.loc[1, target_value]
        else:
            return crosstab.loc[0, target_value]

    def get_freq(self, target_value, privileged=None, predictions=False):
        """
        Returns the frequency of a specified target value. 

        Parameters
        ----------
        target_value:
            Specific value of the target
        privileged (bool, default None): 
            Boolean prescribing whether to
            condition this metric on the `privileged_groups`, if `True`, or
            the `unprivileged_groups`, if `False`. Defaults to `None`
            meaning this metric is computed over the entire dataset.
        predictions: bool
            Whether the frequency is based on predictions or real labels
            
        Returns
        -------
        float
            Frequency of a specified target value
        """
        n_total = self.num_instances(privileged=privileged)
        
        if not predictions:
            n = self.num_spec_value(
                target_value=target_value, privileged=privileged)
        else:
            n = self.num_spec_value(
                target_value=target_value, privileged=privileged, predictions=True)

        return n / n_total

    def base_rate(self, target_value, privileged=None):
        """
        Compute the base rate, :math:`Pr(Y = 1) = P/(P+N)`, optionally
        conditioned on protected attributes.

        Parameters
        ----------
        target_value:
            Specific label value for which it will compute the metric
        privileged (bool, optional):
                Boolean prescribing whether to
                condition this metric on the `privileged_groups`, if `True`, or
                the `unprivileged_groups`, if `False`. Defaults to `None`
                meaning this metric is computed over the entire dataset.

        Returns
        -------
        float:
            Base rate
        """
        return (self.num_spec_value(target_value=target_value, privileged=privileged, predictions=True)
                / self.num_instances(privileged=privileged))

    def false_positive_rate(self, target_value, privileged=None):
        """
        Returns the ratio of true positives to positive examples in the
        dataset, :math:`TPR = TP/P`, optionally conditioned on protected
        attributes.
        Here negatives values correspond to :math:`y <> v`
        So false positives values correspond to :math:`y <> v\ \&\ \hat{y} = v`

        Parameters
        ----------
        target_value:
            Specific label value for which it will compute the metric
        privileged (bool, optional):
                Boolean prescribing whether to
                condition this metric on the `privileged_groups`, if `True`, or
                the `unprivileged_groups`, if `False`. Defaults to `None`
                meaning this metric is computed over the entire dataset.

        Returns
        -------
        float:
            False positive rate
        """
        other_values = [v for v in list(set(self.preds)) if v != target_value]

        FP = self.performances(target_value=target_value,
                               privileged=privileged)['FP']
        N = 0
        for v in other_values:
            N += self.num_spec_value(target_value=v, privileged=privileged)
        if N == 0:
            return 0
        return FP / N

    def true_positive_rate(self, target_value, privileged=None):
        """
        Returns the ratio of true positives to positive examples in the
        dataset, :math:`TPR = TP/P`, optionally conditioned on protected
        attributes.
        Here positives values correspond to :math:`y = v`
        So true positives values correspond to :math:`y = v\ \&\ \hat{y} = v`

        Parameters
        ----------
        target_value:
            Specific label value for which it will compute the metric
        privileged (bool, optional):
                Boolean prescribing whether to
                condition this metric on the `privileged_groups`, if `True`, or
                the `unprivileged_groups`, if `False`. Defaults to `None`
                meaning this metric is computed over the entire dataset.
        
        Returns
        -------
        float:
            True positive rate
        """
        TP = self.performances(target_value=target_value,
                               privileged=privileged)['TP']
        P = self.num_spec_value(
            target_value=target_value, privileged=privileged)
        if P == 0:
            return 0
        return TP / P

    def disparate_impact(self, target_value):
        r"""
        Computes the disparate impact for a specific label value

        .. math::
           \frac{Pr(\hat{Y} = v | D = \text{unprivileged})}
           {Pr(\hat{Y} = v | D = \text{privileged})}

        code source inspired from aif360_

        .. _aif360: https://aif360.readthedocs.io/en/latest/_modules/aif360/metrics/classification_metric.html

        Parameters
        ----------
        target_value:
            Specific label value for which it will compute the metric

        Returns
        -------
        float:
            Disparate impact bias metric
        """
        return self.ratio(self.base_rate, target_value=target_value)

    def statistical_parity_difference(self, target_value):
        r"""
        Computes the statistical parity difference for a specific label value

        .. math::
           Pr(\hat{Y} = v | D = \text{unprivileged})
           - Pr(\hat{Y} = v | D = \text{privileged})

        code source inspired from aif360_

        .. _aif360: https://aif360.readthedocs.io/en/latest/_modules/aif360/metrics/classification_metric.html

        Parameters
        ----------
        target_value:
            Specific label value for which it will compute the metric

        Returns
        -------
        float:
            Statistical parity difference bias metric
        """
        return self.difference(self.base_rate, target_value=target_value)

    def equal_opportunity_difference(self, target_value):
        r"""
        Computes the statistical parity difference for a specific label value
        
        :math:`TPR_{D = \text{unprivileged}} - TPR_{D = \text{privileged}}`

        code source from aif360_

        .. _aif360: https://aif360.readthedocs.io/en/latest/_modules/aif360/metrics/classification_metric.html

        Parameters
        ----------
        target_value:
            Specific label value for which it will compute the metric

        Returns
        -------
        float:
            Equal opportunity difference bias metric
        """
        return self.difference(self.true_positive_rate, target_value=target_value)

    def average_abs_odds_difference(self, target_value):
        r"""
        Average of absolute difference in FPR and TPR for unprivileged and
        privileged groups:

        .. math::

           \frac{1}{2}\left[|FPR_{D = \text{unprivileged}} - FPR_{D = \text{privileged}}|
           + |TPR_{D = \text{unprivileged}} - TPR_{D = \text{privileged}}|\right]

        A value of 0 indicates equality of odds.

        code source from aif360_

        .. _aif360: https://aif360.readthedocs.io/en/latest/_modules/aif360/metrics/classification_metric.html

        Parameters
        ----------
        target_value:
            Specific label value for which it will compute the metric

        Returns
        -------
        float:
            Average of absolute difference bias metric
        """
        return 0.5 * (np.abs(self.difference(self.false_positive_rate, target_value=target_value))
                      + np.abs(self.difference(self.true_positive_rate, target_value=target_value)))

    def theil_index(self, target_value):
        r"""
        Theil index or Generalized entropy index (with $\alpha=1$) is proposed as a unified individual and
        group fairness measure.
        
        With :math:`b_i = \hat{y}_i - y_i + 1`:

        .. math:: 

            \frac{1}{n}\sum_{i=1}^n\frac{b_{i}}{\mu}\ln\frac{b_{i}}{\mu}
            
        code source from aif360_

        .. _aif360: https://aif360.readthedocs.io/en/latest/_modules/aif360/metrics/classification_metric.html

        Parameters
        ----------
        target_value:
            Specific label value for which it will compute the metric

        Returns
        -------
        float:
            Theil index bias metric
        """
        y_pred = self.preds
        y_true = self.labels
        y_pred = (y_pred == target_value).astype(np.float64)
        y_true = (y_true == target_value).astype(np.float64)
        b = 1 + y_pred - y_true

        return np.mean(np.log((b / np.mean(b))**b) / np.mean(b))

    def compute_model_bias_metrics(self):
        """
        Computes automaticaly all dataset bias metrics for this
        protected attribute.
        """
        metrics_dict = {
            'Disparate impact': self.disparate_impact,
            'Statistical parity difference': self.statistical_parity_difference,
            'Equal opportunity difference': self.equal_opportunity_difference,
            'Average abs odds difference': self.average_abs_odds_difference,
            'Theil index': self.theil_index,
        }

        self.compute_bias_metrics(metrics_dict=metrics_dict)
